import logging
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s',
                    level=logging.INFO)

import numpy as np
import pylab as pl
from nacl.dataset import TrainingDataset
from nacl.models.salt import SALT2Like
from nacl.models.constraints import SALT2LikeConstraints, solve_constraints
from nacl.models import constraints
from nacl.models.regularizations import NaClSplineRegularization
from nacl.models import variancemodels
from nacl import minimize
import helpers

tds, model = helpers.generate_dataset(1000, seed=42, string_ids=False,
                                      compress=False)
p_truth = model.pars.copy()


# def test_lightcurve_fit(fit=True, linesearch=False, start_from_truth=False):
#     """fit the light curves, not the model
#     """
#     # re-init model parameters
#     model.pars.full[:] = p_truth.full[:]

#     # fix everything but the SN parameters of interest
#     model.pars.fix()
#     for block_name in ['X0', 'X1', 'col', 'tmax']:
#         model.pars[block_name].release()

#     # tds_lc = TrainingDataset(tds.sn_data, tds.lc_data,
#     #                          load_filters=True)
#     # model_lc = SALT2Like(tds_lc, init_fr)

#     valid = tds.spec_data.valid.copy()
#     tds.spec_data.valid[:] = 0

#     # scramble the initial state
#     if not start_from_truth:
#         sz = len(model.pars['X0'].free)
#         model.pars['X0'].free *= np.random.normal(size=sz, loc=1., scale=0.1)
#         model.pars['X1'].free += np.random.normal(size=sz, scale=1.)
#         model.pars['col'].free += np.random.normal(size=sz, scale=0.3)
#         model.pars['tmax'].free += np.random.normal(size=sz, scale=1.)

#     wres = minimize.WeightedResiduals(model)
#     chi2 = minimize.LogLikelihood(wres)
#     minz = minimize.Minimizer(chi2)
#     if fit:
#         res = minz.minimize(model.pars.free, linesearch=linesearch)
#     else:
#         res = None

#     return res, minz


def test_model_fit(start_from_truth=True, plot=True):
    """
    """
    model.pars.release()
    model.pars.full[:] = p_truth.full[:]
    model.init_from_training_dataset()

    if not start_from_truth:
        sz = len(model.pars['X0'].free)
        model.pars['X0'].free *= np.random.normal(size=sz, loc=1., scale=0.1)
        model.pars['X1'].free += np.random.normal(size=sz, scale=1.)
        model.pars['col'].free += np.random.normal(size=sz, scale=0.2)
        model.pars['tmax'].free += np.random.normal(size=sz, scale=1.)

    model.pars.fix()
    for block_name in ['X0', 'X1', 'col', 'tmax']:
        model.pars[block_name].release()
    ll = minimize.LogLikelihood2(model, force_default_spgemm=False)
    minz = minimize.Minimizer(ll)
    res = minz.minimize_lm(model.pars.free, max_iter=1000,
                           max_attempts=40, accept=10., reject=10.,
                           lamb=1.E-6)


    res['minz'] = minz
    res['truth'] = p_truth

    if plot:
        p_rec = model.pars
        fig, axes = pl.subplots(nrows=2, ncols=2, figsize=(8,8), sharex=True)
        dp = p_rec['X0'].full - p_truth['X0'].full
        axes[0,0].plot(tds.sn_data.z, dp[tds.sn_data.sn_index],'k.')
        axes[0,0].set_title('X0')
        axes[0,0].set_ylabel('$X_0^{rec} - X_0^{true}$')

        dp = p_rec['X1'].full-p_truth['X1'].full
        axes[0,1].plot(tds.sn_data.z, dp[tds.sn_data.sn_index], 'k.')
        axes[0,1].set_title('X1')
        axes[0,1].set_ylabel('$X_1^{rec} - X_1^{true}$')

        dp = p_rec['col'].full-p_truth['col'].full
        axes[1,0].plot(tds.sn_data.z, dp[tds.sn_data.sn_index], 'k.')
        axes[1,0].set_title('$col$')
        axes[1,0].set_ylabel('$col^{rec} - col^{true}$')

        dp = p_rec['tmax'].full-p_truth['tmax'].full
        axes[1,1].plot(tds.sn_data.z, dp[tds.sn_data.sn_index], 'k.')
        axes[1,1].set_title('$t_{max}$')
        axes[1,1].set_ylabel('$t_{max}^{rec} - t_{max}^{true}$')


    return res


def test_model_training_linear_constraints(start_from_truth=True, plot=True):
    """
    """
    model.pars.release()
    model.pars.full[:] = p_truth.full[:]
    model.init_from_training_dataset()

    if not start_from_truth:
        sz = len(model.pars['X0'].free)
        model.pars['X0'].free *= np.random.normal(size=sz, loc=1., scale=0.1)
        model.pars['X1'].free += np.random.normal(size=sz, scale=1.)
        model.pars['col'].free += np.random.normal(size=sz, scale=0.2)
        model.pars['tmax'].free += np.random.normal(size=sz, scale=1.)

    model.pars.fix()
    for block_name in ['X0', 'X1', 'col', 'tmax', 'M0', 'M1', 'CL']:
        model.pars[block_name].release()
        
    cons = constraints.salt2like_linear_constraints(model, mu=1.E6) # was 1.E6
    reg = NaClSplineRegularization(model, to_regularize=['M0', 'M1'], mu=1.) # was 1.E-6

    ll = minimize.LogLikelihood2(model, reg=[reg], cons=[cons],
                                 phot_error_pedestal=0.,
                                 force_default_spgemm=False)
    minz = minimize.Minimizer(ll)

    model.pars['X1'].full[:] -= model.pars['X1'].full.mean()
    model.pars['col'].full[:] -= model.pars['col'].full.mean()

    res = minz.minimize_lm(model.pars.free, max_iter=1000,
                           max_attempts=40, accept=10., reject=2.,
                           lamb=1.E+3, 
                           diag_charge='levenberg')

    # res = minz.minimize(model.pars.free)

    res['minz'] = minz
    res['truth'] = p_truth

    if plot:
        p_rec = model.pars
        fig, axes = pl.subplots(nrows=2, ncols=2, figsize=(8,8), sharex=True)
        dp = p_rec['X0'].full - p_truth['X0'].full
        axes[0,0].plot(tds.sn_data.z, dp[tds.sn_data.sn_index],'k.')
        axes[0,0].set_title('X0')
        axes[0,0].set_ylabel('$X_0^{rec} - X_0^{true}$')

        dp = p_rec['X1'].full-p_truth['X1'].full
        axes[0,1].plot(tds.sn_data.z, dp[tds.sn_data.sn_index], 'k.')
        axes[0,1].set_title('X1')
        axes[0,1].set_ylabel('$X_1^{rec} - X_1^{true}$')

        dp = p_rec['col'].full-p_truth['col'].full
        axes[1,0].plot(tds.sn_data.z, dp[tds.sn_data.sn_index], 'k.')
        axes[1,0].set_title('$col$')
        axes[1,0].set_ylabel('$col^{rec} - col^{true}$')

        dp = p_rec['tmax'].full-p_truth['tmax'].full
        axes[1,1].plot(tds.sn_data.z, dp[tds.sn_data.sn_index], 'k.')
        axes[1,1].set_title('$t_{max}$')
        axes[1,1].set_ylabel('$t_{max}^{rec} - t_{max}^{true}$')


    return res


def test_model_training_linear_constraints_error_snake(tds, start_from_truth=False,
                                                       plot=True):
    """
    """
    model = SALT2Like(tds,
                      init_from_salt2_file='salt2.npz',
                      normalization_band_name='SWOPE::B',
                      error_snake_model_type=variancemodels.SimpleErrorSnake)
    model.init_from_training_dataset()
    # model.init_pars()
    # model.pars.release()
    # model.pars.full[:] = p_truth.full[:]
    # model.pars['sigma_snake'].full[:] = 0.05

    if not start_from_truth:
        sz = len(model.pars['X0'].free)
        model.pars['X0'].free *= np.random.normal(size=sz, loc=1., scale=0.1)
        model.pars['X1'].free += np.random.normal(size=sz, scale=1.)
        model.pars['col'].free += np.random.normal(size=sz, scale=0.2)
        model.pars['tmax'].free += np.random.normal(size=sz, scale=1.)

    model.pars['X1'].full[:] -= model.pars['X1'].full.mean()
    model.pars['col'].full[:] -= model.pars['col'].full.mean()

    # first, fit the model
    model.pars.fix()
    for block_name in ['X0', 'X1', 'col', 'tmax',
                       'M0', 'M1', 'CL']:
        model.pars[block_name].release()

    cons = constraints.salt2like_linear_constraints(model, mu=1.E6)
    reg = NaClSplineRegularization(model, to_regularize=['M0', 'M1'], mu=1.) # was 1.E-6

    ll = minimize.LogLikelihood2(model, reg=[reg], cons=[cons],
                                 variance_model=None, # model.error_snake_model,
                                 force_default_spgemm=False)
    minz = minimize.Minimizer(ll)

    model.pars['X1'].full[:] -= model.pars['X1'].full.mean()
    model.pars['col'].full[:] -= model.pars['col'].full.mean()
    model.pars['sigma_snake'].full[:] = 0.05
    res = minz.minimize_lm(model.pars.free, max_iter=100,
                           max_attempts=40, accept=10., reject=2.,
                           lamb=1.E+3, dchi2_stop=0.01,
                           model='supernodal',
                           diag_charge='levenberg')

    # then, fit the error model only
    model.pars.fix()
    for block_name in ['sigma_snake']:
        model.pars[block_name].release()
    model.pars['sigma_snake'].full[:] = 0.05
    ll = minimize.LogLikelihood2(model, reg=[reg], cons=[cons],
                                 variance_model=model.error_snake_model,
                                 force_default_spgemm=False)
    minz = minimize.Minimizer(ll)

    res = minz.minimize_lm(model.pars.free, max_iter=100,
                           max_attempts=40., accept=10., reject=2.,
                           lamb=1.E+3, dchi2_stop=0.01,
                           mode='supernodal',
                           diag_charge='levenberg')
    
    # finally, release all the parameters
    model.pars.fix()
    for block_name in ['X0', 'X1', 'col', 'tmax',
                       'M0', 'M1', 'CL', 
                       'sigma_snake']:
        model.pars[block_name].release()
    ll = minimize.LogLikelihood2(model, reg=[reg], cons=[cons],
                                 variance_model=model.error_snake_model,
                                 force_default_spgemm=False)
    minz = minimize.Minimizer(ll)
    res = minz.minimize_lm(model.pars.free, max_iter=100,
                           max_attempts=40., accept=10., reject=2.,
                           lamb=1.E+3, dchi2_stop=0.01,
                           mode='supernodal',
                           diag_charge='levenberg')
    res['minz'] = minz
    res['truth'] = p_truth

    if plot:
        p_rec = model.pars
        fig, axes = pl.subplots(nrows=2, ncols=2, figsize=(8,8), sharex=True)
        dp = p_rec['X0'].full - p_truth['X0'].full
        axes[0,0].plot(tds.sn_data.z, dp[tds.sn_data.sn_index],'k.')
        axes[0,0].set_title('X0')
        axes[0,0].set_ylabel('$X_0^{rec} - X_0^{true}$')

        dp = p_rec['X1'].full-p_truth['X1'].full
        axes[0,1].plot(tds.sn_data.z, dp[tds.sn_data.sn_index], 'k.')
        axes[0,1].set_title('X1')
        axes[0,1].set_ylabel('$X_1^{rec} - X_1^{true}$')

        dp = p_rec['col'].full-p_truth['col'].full
        axes[1,0].plot(tds.sn_data.z, dp[tds.sn_data.sn_index], 'k.')
        axes[1,0].set_title('$col$')
        axes[1,0].set_ylabel('$col^{rec} - col^{true}$')

        dp = p_rec['tmax'].full-p_truth['tmax'].full
        axes[1,1].plot(tds.sn_data.z, dp[tds.sn_data.sn_index], 'k.')
        axes[1,1].set_title('$t_{max}$')
        axes[1,1].set_ylabel('$t_{max}^{rec} - t_{max}^{true}$')


    return res






#     # nphot = len(tds.lc_data.flux)
#     # noise = np.random.normal(scale=np.abs(0.05 * tds.lc_data.flux), size=nphot)
#     # tds.lc_data.flux += noise

#     # we start with a simple light curve fit
#     # to do this, we extract just the light curves
#     # from the training dataset
#     lc_tds = TrainingDataset(sne=tds.sn_data.nt, lc_data=tds.lc_data.nt)
#     lc_model = SALT2Like(lc_tds, init_from_salt2_file='salt2.npz')
#     lc_model.init_from_training_dataset()

#     # scramble the initial parameters
#     if not start_from_truth:
#         sz = len(lc_model.pars['X0'].free)
#         lc_model.pars['X0'].free *= np.random.normal(size=sz, loc=1., scale=0.1)
#         lc_model.pars['X1'].free += np.random.normal(size=sz, scale=1.)
#         lc_model.pars['col'].free += np.random.normal(size=sz, scale=0.3)
#         lc_model.pars['tmax'].free += np.random.normal(size=sz, scale=1.)

#     lc_model.pars.fix()
#     lc_model.pars['sigma_snake'].full[:] = 0.1
#     for block_name in ['X0', 'X1', 'col', 'tmax']:  # 'sigma_snake']:
#         lc_model.pars[block_name].release()
#     chi2 = minimize.LogLikelihood2(lc_model,
#                                    # variance_model=lc_model.error_snake,
#                                    force_default_spgemm=False)
#     minz = minimize.Minimizer(chi2)
#     res = minz.minimize(lc_model.pars.free)  # , mode='simplicial', beta=1.E-6)

#     # and initialize the full model with the results of the light curve fit
#     for block_name in ['X0', 'X1', 'col', 'tmax']:
#         model.pars[block_name].full[:] = lc_model.pars[block_name].full[:]
#     # then, release the other parameters and start the real fit
#     model.pars['sigma_snake'] = 0.05
#     model.pars.fix()
#     model.pars['sigma_snake'].full[:] = 0.05
#     for block_name in ['X0', 'X1', 'col', 'tmax',
#                        'M0', 'M1', 'CL',
#                        'sigma_snake'
#                        # 'SpectrumRecalibration'
#                        ]:
#         model.pars[block_name].release()
#         # nrecal = len(model.pars['SpectrumRecalibration'].full)
#         # model.pars['SpectrumRecalibration'].full[:] += np.random.uniform(-1.E-6, 1.E-6, size=nrecal)

#     # at first order, make sure that the constraints are respected
#     # model.pars['X1'].full -= model.pars['X1'].full.mean()
#     # model.pars['X1'].full /= model.pars['X1'].full.std()
#     # model.pars['col'].full -= model.pars['col'].full.mean()
#     cons = SALT2LikeConstraints(model,
#                                 active={'M0': 10**(-0.4 * (30-19.5)),
#                                         'dM0': 0.,
#                                         'M1': 0.,
#                                         'dM1': 0.,
#                                         'col': 0.,
#                                         'X1': 0.,
#                                         'X1_var': 1.}, mu=1.E6)
#     reg = NaClSplineRegularization(model, to_regularize=['M0', 'M1'],
#                                    order=0, mu=1.)
#     chi2 = minimize.LogLikelihood2(model, reg=[reg], cons=[cons],
#                                    force_default_spgemm=False,
#                                    variance_model=model.error_snake
#                                    )
#     minz = minimize.Minimizer(chi2)
#     # dx = solve_constraints(cons, model.pars)
#     # model.pars.free += dx
#     # model.pars['X1'].full /= model.pars['X1'].full.std()
#     # print('cons=', cons(model.pars.free))
#     res = minz.minimize(model.pars.free)

#     res['lc_model'] = lc_model
#     res['model'] = model
#     res['cons'] = cons
#     res['reg'] = reg
#     res['chi2'] = chi2

#     return res





def test_model_fit_no_constraints():
    """Simple model fit, some regularization, no constraints,
    just fixing a couple of SNe to break the degeneracies.
    """
    model.pars.release()
    model.pars.full[:] = p_truth.full[:]

    model.pars.fix()
    for block_name in ['X0', 'X1', 'col', 'tmax', 'M0', 'M1', 'CL']:
        model.pars[block_name].release()
    for sn in [0,1,2,3]:
        model.pars['X0'].fix(sn)
        model.pars['X1'].fix(sn)
        model.pars['col'].fix(sn)
        model.pars['tmax'].fix(sn)

    reg = NaClSplineRegularization(model, to_regularize=['M0', 'M1'],
                                   order=1, mu=1.E-0)
    wres = minimize.WeightedResiduals(model)
    chi2 = minimize.LogLikelihood(wres, reg=[reg])
    minz = minimize.Minimizer(chi2)
    res = minz(model.pars.free, linesearch=False)

    # res['cons'] = cons
    res['reg'] = reg
    res['chi2'] = chi2

    return res


def test_training_before_after(start_from_truth=False):
    """
    """
    model.pars.release()
    model.pars.full[:] = p_truth.full[:]

    flux = tds.spec_data.flux
    fluxerr = tds.spec_data.fluxerr
    tds.spec_data.fluxerr[:] = np.sqrt(fluxerr**2 + 0.1**2 * flux**2)

    # we may want to scramble the initial parameters
    if not start_from_truth:
        sz = len(model.pars['X0'].free)
        model.pars['X0'].free *= np.random.normal(size=sz, loc=1., scale=0.1)
        model.pars['X1'].free += np.random.normal(size=sz, scale=1.)
        model.pars['col'].free += np.random.normal(size=sz, scale=0.3)
        model.pars['tmax'].free += np.random.normal(size=sz, scale=1.)

    # first, we fit the X0,X1,c,tmax only
    # (on the full dataset)
    model.pars.fix()
    for block_name in ['X0', 'X1', 'col', 'tmax']:
        model.pars[block_name].release()
    chi2 = minimize.LogLikelihood2(model, force_default_spgemm=False)
    minz = minimize.Minimizer(chi2)
    p = minz.minimize(model.pars.free)
    p_fit1 = p.copy()
    print('chi2 after fit1: ', chi2(p_fit1['pars'].free))

    # then, we fit the full model
    for block_name in ['X0', 'X1', 'col', 'tmax', 'M0', 'CL']:
        model.pars[block_name].release()
    cons = SALT2LikeConstraints(model,
                                active={'M0': 10**(-0.4 * (30-19.5)),
                                        'dM0': 0.,
                                        'M1': 0.,
                                        'dM1': 0.,
                                        'col': 0.,
                                        'X1': 0.,
                                        'X1_var': 1.}, mu=1.)
    reg = NaClSplineRegularization(model, to_regularize=['M0', 'M1'],
                                   order=0, mu=1.E-6)
    chi2 = minimize.LogLikelihood2(model, reg=[reg], cons=[cons],
                                   force_default_spgemm=False,
                                   )
    print('chi2 before fit2: ', chi2(model.pars.free))

    minz = minimize.Minimizer(chi2)
    p = minz.minimize(model.pars.free)
    p_fit2 = p.copy()
    print('chi2 after fit2: ', chi2(p_fit2['pars'].free))

    ret = {}
    ret['model'] = model
    ret['p_fit1'] = p_fit1
    ret['p_fit2'] = p_fit2

    return ret


def test_training_fit_new_likelihood(start_from_truth=True):
    """
    """
    model.pars.release()
    model.pars.full[:] = p_truth.full[:]
    model.init_from_training_dataset()

    # nphot = len(tds.lc_data.flux)
    # noise = np.random.normal(scale=np.abs(0.05 * tds.lc_data.flux), size=nphot)
    # tds.lc_data.flux += noise

    # we start with a simple light curve fit
    # to do this, we extract just the light curves
    # from the training dataset
    lc_tds = TrainingDataset(sne=tds.sn_data.nt, lc_data=tds.lc_data.nt)
    lc_model = SALT2Like(lc_tds, init_from_salt2_file='salt2.npz')
    lc_model.init_from_training_dataset()

    # scramble the initial parameters
    if not start_from_truth:
        sz = len(lc_model.pars['X0'].free)
        lc_model.pars['X0'].free *= np.random.normal(size=sz, loc=1., scale=0.1)
        lc_model.pars['X1'].free += np.random.normal(size=sz, scale=1.)
        lc_model.pars['col'].free += np.random.normal(size=sz, scale=0.3)
        lc_model.pars['tmax'].free += np.random.normal(size=sz, scale=1.)

    lc_model.pars.fix()
    lc_model.pars['sigma_snake'].full[:] = 0.1
    for block_name in ['X0', 'X1', 'col', 'tmax']:  # 'sigma_snake']:
        lc_model.pars[block_name].release()
    chi2 = minimize.LogLikelihood2(lc_model,
                                   # variance_model=lc_model.error_snake,
                                   force_default_spgemm=False)
    minz = minimize.Minimizer(chi2)
    res = minz.minimize(lc_model.pars.free)  # , mode='simplicial', beta=1.E-6)

    # and initialize the full model with the results of the light curve fit
    for block_name in ['X0', 'X1', 'col', 'tmax']:
        model.pars[block_name].full[:] = lc_model.pars[block_name].full[:]
    # then, release the other parameters and start the real fit
    model.pars['sigma_snake'] = 0.05
    model.pars.fix()
    model.pars['sigma_snake'].full[:] = 0.05
    for block_name in ['X0', 'X1', 'col', 'tmax',
                       'M0', 'M1', 'CL',
                       'sigma_snake'
                       # 'SpectrumRecalibration'
                       ]:
        model.pars[block_name].release()
        # nrecal = len(model.pars['SpectrumRecalibration'].full)
        # model.pars['SpectrumRecalibration'].full[:] += np.random.uniform(-1.E-6, 1.E-6, size=nrecal)

    # at first order, make sure that the constraints are respected
    # model.pars['X1'].full -= model.pars['X1'].full.mean()
    # model.pars['X1'].full /= model.pars['X1'].full.std()
    # model.pars['col'].full -= model.pars['col'].full.mean()
    cons = SALT2LikeConstraints(model,
                                active={'M0': 10**(-0.4 * (30-19.5)),
                                        'dM0': 0.,
                                        'M1': 0.,
                                        'dM1': 0.,
                                        'col': 0.,
                                        'X1': 0.,
                                        'X1_var': 1.}, mu=1.E6)
    reg = NaClSplineRegularization(model, to_regularize=['M0', 'M1'],
                                   order=0, mu=1.)
    chi2 = minimize.LogLikelihood2(model, reg=[reg], cons=[cons],
                                   force_default_spgemm=False,
                                   variance_model=model.error_snake
                                   )
    minz = minimize.Minimizer(chi2)
    # dx = solve_constraints(cons, model.pars)
    # model.pars.free += dx
    # model.pars['X1'].full /= model.pars['X1'].full.std()
    # print('cons=', cons(model.pars.free))
    res = minz.minimize(model.pars.free)

    res['lc_model'] = lc_model
    res['model'] = model
    res['cons'] = cons
    res['reg'] = reg
    res['chi2'] = chi2

    return res




def test_training_fit(start_from_truth=True):
    """fit the light curves and a block of the model
    """
    model.pars.release()
    model.pars.full[:] = p_truth.full[:]

    # we start with a simple light curve fit
    # to do this, we extract just the light curves
    # from the training dataset
    lc_tds = TrainingDataset(sne=tds.sn_data.nt, lc_data=tds.lc_data.nt)
    lc_model = SALT2Like(lc_tds, init_from_salt2_file='salt2.npz')
    lc_model.init_from_training_dataset()

    # scramble the initial parameters
    if not start_from_truth:
        sz = len(lc_model.pars['X0'].free)
        lc_model.pars['X0'].free *= np.random.normal(size=sz, loc=1., scale=0.1)
        lc_model.pars['X1'].free += np.random.normal(size=sz, scale=1.)
        lc_model.pars['col'].free += np.random.normal(size=sz, scale=0.3)
        lc_model.pars['tmax'].free += np.random.normal(size=sz, scale=1.)

    lc_model.pars.fix()
    for block_name in ['X0', 'X1', 'col', 'tmax']:
        lc_model.pars[block_name].release()
    wres = minimize.WeightedResiduals(lc_model)
    chi2 = minimize.LogLikelihood(wres)
    minz = minimize.Minimizer(chi2)
    res = minz.minimize(lc_model.pars.free)

    # and initialize the full model with the results of the light curve fit
    for block_name in ['X0', 'X1', 'col', 'tmax']:
        model.pars[block_name].full[:] = lc_model.pars[block_name].full[:]
    # then, release the other parameters and start the real fit
    model.pars.fix()
    for block_name in ['X0', 'X1', 'col', 'tmax',
                       'M0', 'M1', 'CL']:  # 'SpectrumRecalibration']:
        model.pars[block_name].release()

    # at first order, make sure that the constraints are respected
    model.pars['X1'].full -= model.pars['X1'].full.mean()
    model.pars['X1'].full /= model.pars['X1'].full.std()
    model.pars['col'].full -= model.pars['col'].full.mean()
    cons = SALT2LikeConstraints(model,
                                active={'M0': 10**(-0.4 * (30-19.5)),
                                        'dM0': 0.,
                                        'M1': 0.,
                                        'dM1': 0.,
                                        'col': 0.,
                                        'X1': 0.,
                                        'X1_var': 1.}, mu=1.E6)
    reg = NaClSplineRegularization(model, to_regularize=['M0', 'M1'],
                                   order=0, mu=1.)
    wres = minimize.WeightedResiduals(model)
    chi2 = minimize.LogLikelihood(wres, reg=[reg], cons=[cons])
    minz = minimize.Minimizer(chi2)
    res = minz.minimize(model.pars.free)

    res['cons'] = cons
    res['reg'] = reg
    res['chi2'] = chi2

    return res


def test_cons_dM0():
    cons = SALT2LikeConstraints(model,
                                active={'dM0': 0.},
                                mu=1.)
    pp = model.pars.copy()

    l = []
    for dx in [-5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5]:
        M0 = pp['M0'].full.copy()
        M0 = np.roll(M0.reshape(-1, 129), dx, axis=0)
        model.pars['M0'].full[:] = M0.ravel()
        l.append((dx, float(cons.Jac @ model.pars.full)))
    l = np.rec.fromrecords(l, names=['dx', 'cons'])
    return l

def test_likelihood_derivatives(start_from_truth=False):
    """
    """
    model.pars.release()
    model.pars.full[:] = p_truth.full[:]

    if not start_from_truth:
        sz = len(model.pars['X0'].free)
        model.pars['X0'].free *= np.random.normal(size=sz, loc=1., scale=0.1)
        model.pars['X1'].free += np.random.normal(size=sz, scale=1.)
        model.pars['col'].free += np.random.normal(size=sz, scale=0.3)
        model.pars['tmax'].free += np.random.normal(size=sz, scale=1.)

    # fix everything but the SN parameters of interest
    # plus one block
    model.pars.fix()
    for block_name in ['X0', 'M0', 'X1', 'col', 'tmax']:
        model.pars[block_name].release()

    model.pars['X1'].full -= model.pars['X1'].full.mean()
    model.pars['X1'].full /= model.pars['X1'].full.std()
    model.pars['col'].full -= model.pars['col'].full.mean()

    cons = SALT2LikeConstraints(model,
                                active={'M0': 10**(-0.4 * (30-19.5)),
                                        'dM0': 0.,
                                        'M1': 0.,
                                        'dM1': 0.,
                                        'col': 0.,
                                        'X1': 0.,
                                        'X1_var': 1.
                                               },
                                mu=1.E6)
    # return cons
    reg = NaClSplineRegularization(model, to_regularize=['M0', 'M1'],
                                   order=0, mu=1.e-6)
    wres = minimize.WeightedResiduals(model)
    chi2 = minimize.LogLikelihood(wres)  # , reg=[reg], cons=[cons])

    Ja, Jn = helpers.check_deriv(chi2, p=model.pars.free)

    return Ja, Jn





"""Log Likelihood
"""

import numpy as np
import scipy.sparse as sparse
import logging

try:
    from sparse_dot_mkl import gram_matrix_mkl, dot_product_mkl
except:
    logging.warning('module: `sparse_dot_mkl` not available')
else:
    logging.info('sparse_dot_mkl found. Building hessian should be faster.')


from saltworks import FitParameters

logging.basicConfig(format="%(asctime)s %(levelname)s %(message)s",
                    level=logging.INFO)



class LogLikelihood:
    """New log likelihood class
    """
    def __init__(self, model, variance_model=None,
                 cons=None, reg=None, priors=None,
                 data=None, error_pedestal=None,
                 force_default_spgemm=False):
        """
        """
        # the log-likelihood constituents
        self.model = model
        self.cons = cons
        self.reg = reg
        self.priors = priors
        self.variance_model = variance_model

        # now, we can create a parameter vector
        self.pars = self.init_pars()

        self.force_default_spgemm = force_default_spgemm

        self.data = data
        if data is not None:
            self.y, self.yerr, self.bads = np.array(data.y), np.array(data.yerr), np.array(data.bads)
        else:
            self.y, self.yerr, self.bads = np.array(model.y), np.array(model.yerr), np.array(model.bads)

        # in general, it is better to fit a real error model but if you are lazy
        # (we are all) you may prepare an error pedestal and add it in
        # quadrature to the measurement errors
        self.error_pedestal = error_pedestal
        self.log = {}

    def init_pars(self):
        """create and initialize full parameter vector
        """
        # retrieve the fitparameter structure from our constituents
        # Note: constraints and regularization objects are not expected
        # to add parameters to the problem.
        struct = self.model.get_struct()
        if self.variance_model is not None:
            struct.extend(self.variance_model.get_struct())
        if self.priors is not None:
            for p in self.priors:
                struct.extend(p.get_struct())

        pars = FitParameters(struct)

        # now, initialize the parameter vector note: constraints and
        # regularization may change the initialization parameters, but they
        # don't have to...
        try:
            self.model.init_pars(pars)
        except:
            self.model.init_pars()
        if self.variance_model is not None and hasattr(self.variance_model, 'init_pars'):
            self.variance_model.init_pars(pars)
        if self.cons is not None:
            for c in self.cons:
                if hasattr(c, 'init_pars'):
                    c.init_pars(pars)
        if self.reg is not None:
            for r in self.reg:
                if hasattr(r, 'init_pars'):
                    r.init_pars(pars)
        if self.priors is not None:
            for r in self.priors:
                if hasattr(r, 'init_pars'):
                    r.init_pars(pars)
        return pars

    def ndof(self):
        """return the degrees of freedom
        """
        nmeas = (~self.bads).sum()
        n_free_pars = len(self.pars.free)
        return int(nmeas - n_free_pars)

    def get_log(self):
        return self.log.copy()

    def __call__(self, p, deriv=False):
        """
        """
        self.pars.free = p
        self.log = {}

        # the ingredients
        main_chi2 = 0.
        cons_penalty = 0.
        reg_penalty = 0.
        bads = self.bads

        # evaluate the model and variance model
        pred, J = _eval_model(self.model, self.pars, jac=deriv)

        # diagonal variance and main chi2
        # total variance is the sum of the measurement errors
        # and the variance model
        var = self.yerr**2
        if self.variance_model:
            vv, J_var  = _eval_model(self.variance_model, self.pars, jac=deriv)
            var += vv

        # we assume here that the user knows what s/he is doing and that s/he
        # has carefully prepare an error pedestal of the correct dimension, to
        # add to the model. We do not check anything here.
        if self.error_pedestal is not None:
            var += self.phot_error_pedestal ** 2

        # the residuals and the chi2
        res = self.y - pred
        sig = np.sqrt(var)
        w_res = res / sig
        # TODO: check: bads !
        self.w_res = w_res
        main_chi2 = (w_res**2)[~bads].sum()

        # the log of the determinant
        if self.variance_model is not None:
            log_det_v = np.log(var[~bads]).sum()
        else:
            log_det_v = 0.

        # and the total log-likelihood before taxes
        llk = main_chi2 + log_det_v

        # evaluate the penalties and add them to the chi2
        # quadratic regularization
        reg_penalty, reg_grad, reg_hess = _eval_penalties(self.reg, self.pars, deriv=deriv)
        llk += reg_penalty

        # quadratic constraints
        cons_penalty, cons_grad, cons_hess = _eval_penalties(self.cons, self.pars, deriv=deriv)
        llk += cons_penalty

        # general priors
        prior_penalty, prior_grad, prior_hess = _eval_penalties(self.priors, self.pars, deriv=deriv)
        llk += prior_penalty

        self.log = {'llk': llk,
                    'main_chi2': main_chi2,
                    'log_det_v': log_det_v,
                    'cons': np.sum(cons_penalty),
                    'reg': np.sum(reg_penalty),
                    'priors': np.sum(prior_penalty)}

        if not deriv:
            return llk

        # OK That was the easy part.
        # Now, we need to compute the derivatives

        # 1. the gradient and hessian have several components
        # first, the two classical ones: J^TWJ and J^TWR
        # J^TWJ is difficult to compute.
        w = 1. / np.sqrt(var)
        N = len(self.y)
        W = sparse.dia_matrix((w, [0]), shape=(N,N))
        # in general we discard the bads only at the last minute.
        # we make an exception for w_J
        w_J = (W @ J)[~bads,:]

        grad = -2. * w_J.T @ w_res[~bads]
        w_J = w_J.tocsr()


        # gram_matrix_mkl implements a multiproc version
        # -- much faster if several cores are available
        if 'gram_matrix_mkl' in globals() and not self.force_default_spgemm:
            logging.debug('hessian: H = J.T @ J (gram_matrix_mkl)')
            # there seem to be a bug in gram_matrix_mkl when one of the matrix
            # dimensions is one. typically refuses to contract (N,1) matrix into
            # a scalar. Hence this workaround
            if len(w_J.data) == 0:
                hess = 2. * w_J.T @ w_J
            elif 1 in w_J.shape:
                hess = 2. * dot_product_mkl(w_J.T, w_J, reorder_output=True)
            else:
                # `reorder_output=True` seems essential here.
                # otherwise, the Newton-Raphson step is wrong, typically
                # by a factor 0.5 ... this is scary, I know...
                # print(f'[DEBUG]  {w_J.shape} {w_J.__class__} {w_J.data}')
                hess = 2. * gram_matrix_mkl(w_J, transpose=False, reorder_output=True).tolil() # was transpose=False !
                # print(f'[DEBUG] {hess.shape} {hess.__class__}')
                row, col = hess.nonzero()
                hess[col,row] = hess[row,col]
        # slow single core version
        else:
            logging.debug('hessian: H = J.T @ J (slow version)')
            hess = 2. * w_J.T @ w_J

        # then, if a variance model has been specified, we need to
        # add its contributions to the gradient and the hessian
        if self.variance_model is not None:
            logging.debug('variance model')
            WW = sparse.dia_matrix((w**2, [0]), shape=(N,N))
            J_var = (WW @ J_var)[~bads, :]

            # gradient
            # TODO: check that bad removal did not introduce a bug here
            mvJ = J_var.tocoo()
            w_res_mvJ = sparse.coo_matrix(
                (mvJ.data * w_res[mvJ.row],
                 (mvJ.row, mvJ.col)),
                shape=mvJ.shape)
            rWdVWr = 1. * w_res_mvJ.T @ w_res

            grad += -1. * rWdVWr
            tr_WdV = 1. * np.array(J_var.T.sum(axis=1)).squeeze()
            grad += tr_WdV

            # hessian
            tr_WdVWdV = 1. * J_var.T.dot(J_var)
            hess += tr_WdVWdV

        # OK. That was the hard part
        # now, the quadratic penalties
        if cons_grad is not None and cons_hess is not None:
            grad += cons_grad
            hess += cons_hess

        if reg_grad is not None and reg_hess is not None:
            grad += reg_grad
            hess += reg_hess

        if prior_grad is not None and prior_hess is not None:
            grad += prior_grad
            hess += prior_hess

        msg = f'chi2={main_chi2:.6e} | log_det_v={log_det_v} | cons='
        msg += f'{cons_penalty:8.6e}'
        msg += ' | reg='
        msg += f'{reg_penalty:.6e}'
        msg += ' | prior='
        msg += f'{prior_penalty:.6e}'
        logging.info(msg)

        return llk, grad, hess


def _eval_model(f, pars, jac):
    """A wrapper around the model, to make its evaluation more generic
    """
    val, J = None, None
    if not jac:
        val = f(pars, jac=False)
    else:
        val, J = f(pars, jac=True)
    return val, J


def _eval_penalties(penalties, pars, deriv, return_log=False):
    """A wrapper to simplify the call to the quadratic penalties
    """
    _log = []
    val, grad, hess = 0., None, None
    if penalties is None:
        return val, grad, hess

    for pen in penalties:
        if pen is None:
            return val, grad, hess
        if not deriv:
            v = pen(pars, deriv=False)
            val += v
        else:
            v, gr, hs = pen(pars, deriv=True)
            val += v
            grad = gr if grad is None else grad + gr
            hess = hs if hess is None else hess + hs
        if return_log:
            _log.append(v)
    if return_log:
        return val, grad, hess, _log
    return val, grad, hess



import nacl.simulations.ideal as ideal
from nacl.models.salt import SALT2Like
from nacl.simulations.flux_simulator import Simulator
from nacl.models.variancemodels import SimpleVarianceModel, ColorScatter
from nacl.minimizers import Minimizer
from nacl.fit import FitModel2D
import numpy as np 


N_SN = 500
MU_REG = 10

####################################### SIMULATION ###################################
# Seed                                                                
SEED_SAMPLING, SEED_NOISE, SEED_INIT = 0, 0, 0

# Create the cadence 
DS = ideal.Generator(n_sn=N_SN, lsst_simu=True)

# Simulate the flux and the measurements error
INIT_FROM_SALT2_FILE = "../data/salt2.npz"
# Color Scatter initialisation
SIGMA_KAPPA_INIT = np.array([0.015, -0.02, 0.05])
CS_FUNC = ColorScatter
# Error Snake value used in simulation
# Error snake initiate at 5%
GAMMA_INIT = 0.05
# Covariance calibration matrix 
COV_CALIB = 1e-6

FS = Simulator(DS,                                        # DataSet
               SALT2Like,                                 # Model function used to simulated
               variance_model_func=SimpleVarianceModel,   # Error Snake used in sim
               gamma_init=GAMMA_INIT,                     # Value of the error snake
               seed=SEED_NOISE,                           # Seed for noise generation                
               sigma_kappa_init=SIGMA_KAPPA_INIT,         # Value of Color scatter used in sim
               eta_covmatrix=COV_CALIB,                   # Value of Calibration Covariance Matrix
               color_scatter_func=CS_FUNC,                # Color scatter mmodel func
               init_from_salt2_file=INIT_FROM_SALT2_FILE) # Model values for sim

####################################### FIT #########################################
## Instantiate Fit
# Create Constraints and Regularization
# as function of free parameters.
fit = FitModel2D(FS.trainingDataset,                            # TrainingDataset
                 model_func=SALT2Like,                          # Model used
                 normalization_band_name='SWOPE::B',            # Model normalization
                 init_from_salt2_file=INIT_FROM_SALT2_FILE,     # Files to initiate the model
                 pars_fix=[],                                   # Parameters that should be fixed
                 color_scatter_func=CS_FUNC,                    # Color Scatter fun
                 mu_reg=MU_REG,                                 # Regularization hyperparameter
                 minimizer=Minimizer)                           # Minimizer function

# degrade SN parameters (multiply by N(1, 0.05))
# degrade error snake  (multiply by N(1, 1))

# select SN parameters
pars_fix = ['M0', 'CL', 'M1', 'SpectrumRecalibration', 'eta_calib', 'kappa_color']
# Fix model parameters
fit.pars_fix = pars_fix
fit.pars_release()

# set kappa and eta to zero
fit.model.pars['eta_calib'].full[:] = 0
fit.model.pars['kappa_color'].full[:] = 0

fit.parameter_degradation(model_coef_init=0.05,
                          var_coef_init=0.1,
                          seed=SEED_INIT)
sigma_kappa = SIGMA_KAPPA_INIT.copy()/2.5 

fit.beta = 1e-10
#####################################################################################
################################################### 1st fit model without Error Snake 
fit(pars_fix=['gamma'],
    sigma_kappa=sigma_kappa,
    g=None,
    eta_covmatrix=COV_CALIB,
    sigma_kappa_fit=True)
sigma_kappa = fit.color_scatter.pars.full[:].copy()

################################################### 2nd Fit only error snake
# Release Error Snake parameter
fit.variance_model.pars.release()
fit(pars_fix=fit.all_pars, g=fit.variance_model.pars.free)

################################################### 3rd Fit : All
# Release model parameters
fit.model.pars.release()
fit(pars_fix=[],
    g=fit.variance_model.pars.free, 
    sigma_kappa=sigma_kappa,
    eta_covmatrix=COV_CALIB,
    sigma_kappa_fit=True)



#!/usr/bin/env python3

#!/usr/bin/env python3

import logging
import time

import numpy as np
import matplotlib.pyplot as plt

from numba import jit, njit

from lemaitre import bandpasses
from nacl.models import helpers
from nacl.models import salt2
from nacl.models.salt2 import lightcurves
from nacl.models.salt2 import spectra
from nacl import TrainingDataset


def load(filename='compressed_lm2', keep=None):
    """
    """
    fl = bandpasses.get_filterlib()

    if 'hd5' in filename:
        tds = TrainingDataset.read_hdf(filename)
    else:
        tds = TrainingDataset.read_parquet(filename, filterlib=fl)

    if keep is not None:
        sne = tds.sn_data.sn_set.tolist()
        for k in keep:
            sne.remove(k)
        tds.kill_sne(sne)
        tds.compress()

    return tds


def eval_deriv(X0, cl, K, zz, calib_corr, cs_corr):
    v_ = X0 * cl * calib_corr * cs_corr * zz
    r = v_[K.row] * K.data
    return r

@njit(parallel=True)
def eval_deriv_jit(X0, cl, k_row, k_data, zz, calib_corr, cs_corr):
    v_ = X0 * cl * calib_corr * cs_corr * zz
    return v_[k_row] * k_data



if __name__ == '__main__':
    tds = load()
    model = salt2.SALT2Like(tds)
    pars = model.init_pars()

    v, J = model(pars, jac=1)

    X0 = model.queue[0].X0
    cl = model.queue[0].cl
    K = model.queue[0].K
    zz = model.queue[0].zz
    calib_corr = model.queue[0].calib_corr
    cs_corr = model.queue[0].cs_corr



import numpy as np 
import pandas as pd
from nacl.models.salt import SALT2Like
import matplotlib.pyplot as plt
from nacl.dataset import TrainingDataset

def cut(data_generator, model_func,init_from_salt2_file="../data/salt2.npz", degree=3, normalization_band_name='SWOPE::B',phase_range=(-20., 50.), wl_range=(2000., 9000.),plot=False):
	r"""
	Function to find the light curve and spectra index outside our model

	Attributes
	----------
	data_generator : nacl.jla.generator or nacl.ideal.generator or nacl.k21.generator
	Cadence and SNe information.
	model_func : nacl.models.salt.SALT2Like
	Model to generated data.
	degree : int
	Degree of the spectral recalibration polynomial function.
	normalization_band_name : str
	Band name for normalization of the model.
	init_from_salt2_file : str
	File salt2, created by script 'nacl.simulation.make_salt2_npz.py'.
	plot : False or True; plot spec/lc with data cut 
	"""

	model = model_func(data_generator.trainingDataset,init_from_salt2_file=init_from_salt2_file, spectrum_recal_degree=degree,normalization_band_name=normalization_band_name,phase_range=phase_range, wl_range=wl_range)


	blue_cutoff, red_cutoff = model.wl_range
	early_cutoff, late_cutoff = model.phase_range

	# for JLA and K21 need cutoff to be less than model def;
	# in order to degrade tmax and fit
	early_cutoff += 5
	late_cutoff -= 5

	lc = data_generator.trainingDataset.lc_data
	sp = data_generator.trainingDataset.spec_data
	snInfo = data_generator.trainingDataset.sn_data

	# lc
	bands = np.unique(lc.nt['band'])
	lambda_range = np.linspace(1500, 11000, int(1e5))
	min_lambda = np.zeros(len(lc.nt['band']))
	max_lambda = np.zeros(len(lc.nt['band']))
	for bd in bands:
		idx = lc.nt['band'] == bd
		ff = model.filter_db.transmission_db[bd].func(lambda_range)
		idxx = np.where(ff > ff.max()*0.01)
		mmin, mmax = lambda_range[idxx[0][0]], lambda_range[idxx[0][-1]]
		min_lambda[idx] = mmin/(1+snInfo.nt.z[np.array([np.where(snInfo.nt.sn==yy)[0][0] for yy in lc.nt[idx].sn])])
		max_lambda[idx] = mmax/(1+snInfo.nt.z[np.array([np.where(snInfo.nt.sn==yy)[0][0] for yy in lc.nt[idx].sn])])
	date_lc = (lc.nt['mjd'] - snInfo.nt.tmax[np.array([np.where(snInfo.nt.sn==yy)[0][0] for yy in lc.nt.sn])]) / (1. + snInfo.nt.z[np.array([np.where(snInfo.nt.sn==yy)[0][0] for yy in lc.nt.sn])])	

	idx_ph_lc = (date_lc > early_cutoff) & (date_lc < late_cutoff)
	idx_wl_lc = (min_lambda > blue_cutoff) & (max_lambda < red_cutoff)
	idx_lc = idx_ph_lc & idx_wl_lc
	
	if plot==True:
		fig, ax1 = plt.subplots(figsize=(8,6), facecolor='w', edgecolor='k')
		
		#ax1.vlines(x=date_lc,ymin=min_lambda,ymax=max_lambda, color='red', zorder=2,ls='--')
		ax1.plot(date_lc,min_lambda,'ro')
		ax1.plot(date_lc,max_lambda,'ro')						
		
		ax1.axvline(x=early_cutoff,color='k',ls='--')	
		ax1.axvline(x=late_cutoff,color='k',ls='--')	
		ax1.axhline(y=blue_cutoff,color='k',ls='--')	
		ax1.axhline(y=red_cutoff,color='k',ls='--')			
		ax1.set_ylabel('Wavelength min max filter [Angs]',fontsize=14,fontweight='bold')						
		ax1.set_xlabel('Phase since max [days]',fontsize=14,fontweight='bold')								
		plt.show()
	# sp
	zspec=pd.merge(pd.DataFrame(sp.nt),pd.DataFrame(snInfo.nt).set_index('index'),how='inner', on=['sn'])['z']
	tspec=pd.merge(pd.DataFrame(sp.nt),pd.DataFrame(snInfo.nt).set_index('index'),how='inner', on=['sn'])['tmax']
	wl_sp = sp.nt['wavelength'] / (1. + zspec)
	idx_wl_sp = (wl_sp > blue_cutoff) & (wl_sp < red_cutoff)
	date_sp = (sp.nt['mjd'] - tspec) / (1. + zspec)
	idx_ph_sp = (date_sp > early_cutoff) & (date_sp < late_cutoff)

	idx_sp = idx_ph_sp & idx_wl_sp
	idx_sp=idx_sp.values

	if plot==True:
		fig, ax1 = plt.subplots(figsize=(8,6), facecolor='w', edgecolor='k')
		
		ax1.plot(date_sp,wl_sp,'ro')		
		ax1.axvline(x=early_cutoff,color='k',ls='--')	
		ax1.axvline(x=late_cutoff,color='k',ls='--')	
		ax1.axhline(y=blue_cutoff,color='k',ls='--')	
		ax1.axhline(y=red_cutoff,color='k',ls='--')			
		ax1.set_ylabel('Wavelength min max filter [Angs]',fontsize=14,fontweight='bold')						
		ax1.set_xlabel('Phase since max [days]',fontsize=14,fontweight='bold')								
		plt.show()	
	
	data_generator.trainingDataset.lc_data.valid=idx_lc.astype(int)
	data_generator.trainingDataset.spec_data.valid=idx_sp.astype(int)
		
	data_generator.trainingDataset.compress()	
	
	return data_generator
	
	
	
	



"""The spectral part of the model


TODO: remove color law evaluation from SpectrumEvalUnitFast
TODO: rename SpectrumEvalUnitFast
"""


import time
import numpy as np
import scipy


class SpectrumEvalUnitFast(object):
    """Evaluate the model for all SN spectra in the training dataset

    This class is one of the two type of "eval units". Given a chunk of the
    training dataset which corresponds the spectral observations, and given the
    :math:`(X_0, X_1, c, t_{max})` parameters of the SN, compute the quantity:

    .. math::

         \frac{1}{1+z} \left[M_0\left(\frac{\lambda}{1+z}, \mathrm{p}\right) + X_1\ M_1\left(\frac{\lambda}{1+z},
         \mathrm{p}\right) \right]\ 10^{0.4\ c\ P(\frac{\lambda}{1+z})}\ s(\lambda_{rec})

    where

    .. math::
        M_{0|1}(\lambda, \mathrm{p}) = \sum_{k\ell} \theta_{k\ell} B_k(\mathrm{p}) B_l(\mathrm{\lambda})

    and

    .. math::
         \mathrm{p} = \frac{t - t_{max}}{1+z}

    and where :math:`R_s(\lambda)` is a polynomial correction which
    absorbs the wavelength-dependent large scale calibration errors
    affecting the spectrum.

    Again, the evaluation reduces to a sparse matrix multiplication.
    """
    def __init__(self, tds, model):
        """Constructor.

        Parameters
        ----------
        tds : nacl.dataset.TrainingDataset
            Data set of photometric and spectroscopic observations.
        model : nacl.models.salt.SALT2Like
            Model.
        """
        self.training_dataset = tds
        self.data = tds.spec_data
        self.model = model

        self.spec_index = self.data.spec_index
        self.sn_index = self.data.sn_index
        # self.z = self.training_dataset.sn_data.z[self.sn_id]
        self.z = self.data.z

        # Look at that later
        self.color_law = model.color_law
        self.pars = model.pars
        self.basis = model.basis

        self.wl_basis_size = len(model.basis.bx)
        self.ph_basis_size = len(model.basis.by)

        # restframe wavelengths
        self.restframe_wl = self.data.wavelength/(1.+self.z)
        #        self.Jl = model.basis.bx.eval(self.restframe_wl).tocsr()

        # and we can connect directly to the global parameters
        self.M0 = model.pars['M0'].full
        self.M1 = model.pars['M1'].full
        self.cl_pars = model.pars['CL'].full

        # recalibration polynomial (one per spectrum, because we adapt to the spectrum wavelength range)
        #        self.recal_func = SpectrumRecalibrationPolynomials(self.training_dataset, self.model,
        #                                                           self.model.recalibration_degree)
        self.recal_func = self.model.recal_func

    def __call__(self, jac=False, debug_mode=False):
        r"""
        Evaluate the model for a all spectra

        Parameters
        ----------
        jac : bool
            if True compute and return the jacobian matrix
        debug_mode : bool
            if true, just return the model components.

        Returns
        -------
        val : numpy.array
            Model evaluations.
        jacobian : scipy.sparse.csr_matrix
            Jacobian matrix (if jac is true).
        """
        t0 = time.perf_counter()
        pars = self.pars

        # sn-related parameters
        x0, x1 = pars['X0'].full[self.sn_index], pars['X1'].full[self.sn_index]
        c, tmax = pars['col'].full[self.sn_index],  pars['tmax'].full[self.sn_index]

        # we need to re-evaluate the basis on the phases, since tmax changes
        restframe_phases = (self.data.mjd-tmax)/(1.+self.z)
        jacobian = self.basis.eval(self.restframe_wl, restframe_phases + self.model.delta_phase).tocsr()

        # model components
        component_0 = jacobian.dot(self.M0)
        component_1 = jacobian.dot(self.M1)
        polynome_color_law, jacobian_color_law = self.color_law(self.restframe_wl, self.cl_pars, jac=jac)
        color_law = np.power(10., 0.4*c*polynome_color_law)
        zz = 1. + self.z

        # recalibration polynomial
        # if we evaluate this, then recal_func must be instantiated
        # if self.recal_func is not None:
        assert self.recal_func is not None
        recal, jacobian_spec_rec = self.recal_func(jac=jac)
        if jacobian_spec_rec is not None:
            jacobian_spec_rec = jacobian_spec_rec.tocoo()
            # don't know what to do with this
            # jacobian_spec_rec.data *= recal[jacobian_spec_rec.row]

        # recal = np.exp(recal)
        pca = (component_0 + x1 * component_1)
        model = pca * color_law * recal / zz

        if debug_mode:
            return component_0, component_1, color_law, recal, model

        self.model.val.append(model)
        if not jac:
            self.model.timing.append(time.perf_counter()-t0)
            return model

        jacobian = jacobian.tocoo()
        # X0 does not appear in the spectral part of the model
        # hence, dmdX0 = 0

        # dMdX1
        jacobian_i, jacobian_j, jacobian_val = self.model.jacobian_i, self.model.jacobian_j, self.model.jacobian_val
        jacobian_i.append(self.data.row)
        jacobian_j.append(self.model.pars['X1'].indexof(self.sn_index))
        jacobian_val.append(component_1 * color_law * recal / zz)

        # dMdc
        jacobian_i.append(self.data.row)
        jacobian_j.append(self.model.pars['col'].indexof(self.sn_index))
        jacobian_val.append(model * 0.4 * np.log(10.) * polynome_color_law)

        # dMdtmax
        # we can gain a little here, by not evaluating the gradient along the wavelength (ddlambda)
        _, deval_phase = self.model.basis.gradient(self.restframe_wl, restframe_phases + self.model.delta_phase)
        deval_phase = deval_phase.tocsr()
        jacobian_i.append(self.data.row)
        jacobian_j.append(self.model.pars['tmax'].indexof(self.sn_index))
        jacobian_val.append(-1. * (deval_phase.dot(self.M0) + x1*deval_phase.dot(self.M1)) * color_law * recal / zz**2)

        # dmdtheta_0
        jacobian_i.append(self.data.row[jacobian.row])
        jacobian_j.append(self.model.pars['M0'].indexof(jacobian.col))
        jacobian_val.append(jacobian.data * color_law[jacobian.row] * recal[jacobian.row] / zz[jacobian.row])

        # dmdtheta_1
        jacobian_i.append(self.data.row[jacobian.row])
        jacobian_j.append(self.model.pars['M1'].indexof(jacobian.col))
        jacobian_val.append(x1[jacobian.row] * jacobian.data * color_law[jacobian.row] *
                            recal[jacobian.row] / zz[jacobian.row])

        # dMdcl (color law)
        jacobian_color_law = scipy.sparse.coo_matrix(jacobian_color_law)
        jacobian_i.append(self.data.row[jacobian_color_law.row])
        jacobian_j.append(self.model.pars['CL'].indexof(jacobian_color_law.col))
        jacobian_val.append(c[jacobian_color_law.row] * 0.4 * np.log(10.) * jacobian_color_law.data *
                            model[jacobian_color_law.row])

        # dMdr (recalibration)
        if jacobian_spec_rec is not None:
            jacobian_i.append(self.data.row[jacobian_spec_rec.row])
            jacobian_j.append(self.model.pars['SpectrumRecalibration'].indexof(jacobian_spec_rec.col))
            jacobian_val.append(jacobian_spec_rec.data * (pca*color_law)[jacobian_spec_rec.row]/zz[jacobian_spec_rec.row])

        if self.model.disable_cache:
            self.model.clear_cache()
        self.model.timing.append(time.perf_counter()-t0)

        return model


class SpectrumRecalibrationPolynomials:
    """A class to manage the spectrum recalibration polynomials

    The photometric calibration of spectra is generally affected by significant
    error modes at large wavelength scales. It is imperative to remove these
    error modes, while preserving the information provided by the spectra at
    small scales (spectral features)

    This is achieved during training by multiplying the spectral predictions of
    the model by a recalibration polynomial specific to each spectrum, function
    of the observation wavelength :math:`\lambda_o`, and of order :math:`N_s =
    3`, common to all spectra:

    .. math::
            s(\lambda_{rec}) = \sum_i^{N_s} s_i \lambda_{rec}^{N_s - i} %quad \mbox{and}
            \quad \lambda_{0} = \frac{\lambda - 5000}{9000 - 2700}

    with:

    .. math::
            \lambda_{rec} = \frac{2 (\lambda_o - \lambda_{max})}{\lambda_{max} - \lambda_{min}+1
    """

    def __init__(self, tds, model, pol_degrees):
        """Constructor

           Parameters
           ----------
            tds : nacl.dataset.TrainingDataset
                Data set of photometric and spectroscopic observations.
            model : nacl.model.salt
                Model.
            pol_degrees : int or indexable structure
                Polynomial degree for each spectrum.
            """
        self.tds = tds
        self.model = model
        self.pol_degrees = pol_degrees
        self.jacobian = None

        self.N = len(self.tds.spec_data)
        o = np.cumsum(np.hstack(([0], self.pol_degrees+1)))
        self.offset, self.n = o[:-1], o[-1]
        self.build_jacobian_matrix()

    def init_pars(self):
        """return a parameter vector initialized such that
            the recalibration polynomials are evaluated to 1
            for each spectra.

            Returns
            -------
            array
                Initiated parameters.
        """
        p = np.zeros(self.n)
        o = np.cumsum(self.pol_degrees+1) - 1
        p[o] = 1.
        return p

    def build_jacobian_matrix(self):
        """
            Create the jacobian matrix where line correspond to all spectral data point and
            the colones to parameters.
            """
        i, j, v = [], [], []

        # easier to write this with the spectra index
        for sp in self.tds.spectra:
            spec_index = sp.spec_index[0]
            lmin, lmax = sp.wavelength.min(), sp.wavelength.max()
            a = 2. / (lmax-lmin)
            b = 1 - 2. * lmax / (lmax-lmin)
            rwl = a * sp.wavelength + b
            deg = self.pol_degrees[spec_index]
            jacobian_spec_rec = scipy.sparse.coo_matrix(np.vander(rwl, deg+1))
            i.append(jacobian_spec_rec.row + sp.slc.start)
            j.append(jacobian_spec_rec.col + self.offset[spec_index])
            v.append(jacobian_spec_rec.data)
        i = np.hstack(i)
        j = np.hstack(j)
        v = np.hstack(v)
        self.jacobian = scipy.sparse.coo_matrix((v, (i, j)), shape=(self.N, self.n)).tocsr()

    def __call__(self, jac=False):
        """Evaluate the recalibration polynomials

        Parameters
        ----------
        jac : bool
          whether the jacobian is needed.

        Returns
        -------
        v : numpy.array
          Recalibration parameter evaluated.
        self.jacobian : None or scipy.sparse.coo_matrix
          Jacobian matrix of the recalibration polynomial functions.
        """
        p_full = self.model.pars['SpectrumRecalibration'].full
        v = self.jacobian.dot(p_full)
        if not jac:
            return v, None
        return v, self.jacobian


class SpectroPhotoEvalUnit(object):

    def __init__(self, tds, model):
        """Constructor.

        Parameters
        ----------
        tds : nacl.dataset.TrainingDataset
            Data set of photometric and spectroscopic observations.
        model : nacl.models.salt.SALT2Like
            Model.
        """
        self.training_dataset = tds
        self.data = tds.spectrophotometric_data
        self.model = model

        self.spec_index = self.data.spec_index
        self.sn_index = self.data.sn_index
        self.z = self.data.z

        # Look at that later
        self.color_law = model.color_law
        self.pars = model.pars
        self.basis = model.basis

        self.wl_basis_size = len(model.basis.bx)
        self.ph_basis_size = len(model.basis.by)

        # restframe wavelengths
        self.restframe_wl = self.data.wavelength/(1.+self.z)
        #        self.Jl = model.basis.bx.eval(self.restframe_wl).tocsr()

        # and we can connect directly to the global parameters
        self.M0 = model.pars['M0'].full
        self.M1 = model.pars['M1'].full
        self.cl_pars = model.pars['CL'].full

    def __call__(self, jac=False, debug_mode=False):
        r"""
        Evaluate the model for a all spectra

        Parameters
        ----------
        jac : bool
            if True compute and return the jacobian matrix
        debug_mode : bool
            if true, just return the model components.

        Returns
        -------
        val : numpy.array
            Model evaluations.
        jacobian : scipy.sparse.csr_matrix
            Jacobian matrix (if jac is true).
        """

        t0 = time.perf_counter()
        pars = self.pars

        # sn-related parameters
        x0, x1 = pars['X0'].full[self.sn_index], pars['X1'].full[self.sn_index]
        c, tmax = pars['col'].full[self.sn_index],  pars['tmax'].full[self.sn_index]
        # c, tmax = pars['col'].full[self.sn_index],  pars['tmax'].full[self.training_dataset.spectrophotometric_data.isn]

        # we need to re-evaluate the basis on the phases, since tmax changes
        restframe_phases = (self.data.mjd-tmax)/(1.+self.z)
        jacobian = self.basis.eval(self.restframe_wl, restframe_phases + self.model.delta_phase).tocsr()

        # norm
        norm = self.model.norm

        # model components
        component_0 = jacobian.dot(self.M0)
        component_1 = jacobian.dot(self.M1)

        polynome_color_law, jacobian_color_law = self.color_law(self.restframe_wl, self.cl_pars, jac=jac)
        color_law = np.power(10., 0.4*c*polynome_color_law)
        zz = 1. + self.z

        pca = (component_0 + x1 * component_1)
        model = x0 * norm * pca * color_law / zz

        self.model.val.append(model)
        if not jac:
            self.model.timing.append(time.perf_counter()-t0)
            return model

        jacobian = jacobian.tocoo()
        # jacobian_spec_rec = jacobian_spec_rec.tocoo()

        jacobian_i, jacobian_j, jacobian_val = self.model.jacobian_i, self.model.jacobian_j, self.model.jacobian_val
        jacobian_i.append(self.data.row)
        jacobian_j.append(self.model.pars['X0'].indexof(self.sn_index))
        # jacobian_val.append((component_0 + x1*component_1) * color_law * recal / zz)
        jacobian_val.append(norm * pca * color_law / zz)

        # dMdX1
        # jacobian_i, jacobian_j, jacobian_val = self.model.jacobian_i, self.model.jacobian_j, self.model.jacobian_val
        jacobian_i.append(self.data.row)
        jacobian_j.append(self.model.pars['X1'].indexof(self.sn_index))
        # jacobian_val.append(x0 * component_1 * color_law * recal / zz)
        jacobian_val.append(x0 * norm * component_1 * color_law / zz)

        # dMdc
        jacobian_i.append(self.data.row)
        jacobian_j.append(self.model.pars['col'].indexof(self.sn_index))
        jacobian_val.append(model * 0.4 * np.log(10.) * polynome_color_law)

        # dMdtmax
        # we can gain a little here, by not evaluating the gradient along the wavelength (ddlambda)
        _, deval_phase = self.model.basis.gradient(self.restframe_wl, restframe_phases + self.model.delta_phase)
        deval_phase = deval_phase.tocsr()
        jacobian_i.append(self.data.row)
        jacobian_j.append(self.model.pars['tmax'].indexof(self.sn_index))
        #jacobian_val.append(-1. *x0* (deval_phase.dot(self.M0) + x1*deval_phase.dot(self.M1)) * color_law * recal / zz**2)
        jacobian_val.append(-1. * x0 * norm * (deval_phase.dot(self.M0) + x1*deval_phase.dot(self.M1)) * color_law / zz**2)

        # dmdtheta_0
        jacobian_i.append(self.data.row[jacobian.row])
        jacobian_j.append(self.model.pars['M0'].indexof(jacobian.col))
        #jacobian_val.append(x0[jacobian.row] *jacobian.data * color_law[jacobian.row] * recal[jacobian.row] / zz[jacobian.row])
        jacobian_val.append(x0[jacobian.row] * norm * jacobian.data * color_law[jacobian.row] / zz[jacobian.row])

        # dmdtheta_1
        jacobian_i.append(self.data.row[jacobian.row])
        jacobian_j.append(self.model.pars['M1'].indexof(jacobian.col))
        #jacobian_val.append(x0[jacobian.row]*x1[jacobian.row] * jacobian.data * color_law[jacobian.row] *
        #                    recal[jacobian.row] / zz[jacobian.row])
        jacobian_val.append(x0[jacobian.row] * x1[jacobian.row] * norm *  jacobian.data * color_law[jacobian.row]/ zz[jacobian.row])

        # dMdcl (color law)
        jacobian_color_law = scipy.sparse.coo_matrix(jacobian_color_law)
        jacobian_i.append(self.data.row[jacobian_color_law.row])
        jacobian_j.append(self.model.pars['CL'].indexof(jacobian_color_law.col))
        jacobian_val.append(c[jacobian_color_law.row] * 0.4 * np.log(10.) * jacobian_color_law.data *
                            model[jacobian_color_law.row])

        if self.model.disable_cache:
            self.model.clear_cache()
        self.model.timing.append(time.perf_counter()-t0)


        return model




# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information
import sphinx_rtd_theme
import sphinx_gallery

project = 'NaCl'
copyright = '2023, Nicolas Regnault, Guy Augarde, Marc Betoule, Seb Bongard'
author = 'Nicolas Regnault, Guy Augarde, Marc Betoule, Seb Bongard'
release = '0.1'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.inheritance_diagram',
    'sphinx.ext.mathjax',
    'sphinxemoji.sphinxemoji',
    'm2r2',
]

templates_path = ['_templates']
exclude_patterns = []



# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output
# html_theme = 'alabaster'
# html_theme = "sphinx_book_theme"
html_theme = 'sphinx_rtd_theme'
html_theme_options = {
    # 'show_toc_level': 2,
    # 'repository_url': 'https://gitlab.in2p3.fr/cosmo/sn-nacl',
    # 'use_repository_button': True,     # add a "link to repository" button
}

html_static_path = ['_static']



#!/usr/bin/env python3

import logging

import numpy as np
import matplotlib.pyplot as plt

from lemaitre import bandpasses
from nacl.models import helpers
from nacl.models import salt2
from nacl.models.salt2 import lightcurves
from nacl.models.salt2 import spectra
from nacl import TrainingDataset
from nacl.loglikelihood import LogLikelihood
from nacl.minimize import Minimizer



def load(filename='snf_projected.hd5', keep=['SNF20051003-004', 'PTF10qyz', 'PTF11kly', 'SNF20080803-000']):
    """
    """
    fl = bandpasses.get_filterlib()

    if 'hd5' in filename:
        tds = TrainingDataset.read_hdf(filename)
    else:
        tds = TrainingDataset.read_parquet(filename, filterlib=fl)

    if keep is not None:
        sne = tds.sn_data.sn_set.tolist()
        for k in keep:
            sne.remove(k)
        tds.kill_sne(sne)
        tds.compress()

    return tds


def plot_spectrum(model, pars, spec=None, spec_index=None):
    """
    """
    tds = model.training_dataset
    if spec is not None:
        idx = tds.spectrophotometric_data.spec == spec
    elif spec_index is not None:
        idx = tds.spectrophotometric_data.spec_index == spec_index
    else:
        return

    ib = tds.spectrophotometric_data.i_basis[idx]
    wl = tds.spectrophotometric_data.wavelength[idx]
    flx = tds.spectrophotometric_data.flux[idx]
    flxerr = tds.spectrophotometric_data.fluxerr[idx]

    print(flx)

    v = model(pars)

    fig, axes = plt.subplots(nrows=2, ncols=1, figsize=(8,8), sharex=True)
    axes[0].errorbar(wl, flx, yerr=flxerr, ls='', marker='.', color='b')
    axes[0].plot(wl, v[idx], 'r+')
    axes[1].plot(wl, flx/v[idx], 'b.')
    print(np.median(flx/v[idx]))





if __name__ == '__main__':
    tds = load(keep=None)
    # quick hack
    n = len(tds.spectrophotometric_data)
    tds.spectrophotometric_data.add_field('valid', np.ones(n).astype(int))

    model = salt2.SALT2Like(tds, wl_grid=np.linspace(2000., 11000., 202))
    # pars = model.init_pars()
    # wl = np.linspace(2200., 9200., 250)
    #    for block_name in ['X0', 'X1', 'col', 'tmax', 'M0', 'M1', 'CL']:
    #        logging.info(f'testing: {block_name}')
    #        test_deriv(model, pars, block_name)


    ll = LogLikelihood(model)
    init_pars = ll.pars.copy()

    ll.pars.fix()
    ll.pars.release('X1')
    mm = Minimizer(ll)
    #mm.minimize(ll.pars.free)

    #plt.figure()
    #plt.plot(init_pars['X0'].full-ll.pars['X0'].full, 'k.')



import numpy as np
import pylab as pl
pl.ion()
import sncosmo
from nacl.models.salt2.salt import SALT2Like
from nacl.dataset import SimTrainingDataset
source = sncosmo.get_source('salt2', version='2.4')
import os
os.environ['KMP_DUPLICATE_LIB_OK']='True'

from lemaitre import bandpasses

def NaCl_vs_sncosmo_simple(bands = ['ztf::g'], params = [{'x0':1, 'x1':0, 'c':0}], dt=0.27, zp=30., zp_band=1.):
    """
    This function compares NaCl lightcurve prediction with sncosmo in a single band with a set of given parameters
    and calculates the relative difference.
    """
    fl = bandpasses.get_filterlib()
    model_sncosmo = sncosmo.Model(source=source)
    t = np.arange(-20, 50, 1)
    for band in bands:
        tds_sim = SimTrainingDataset( [band], n_spectra=0, n_phot_spectra=0, zp=zp, filterlib=fl )
        model_sim = SALT2Like(tds_sim, basis_knots=[700,70])
        pars = model_sim.init_pars()
        for p in params:
            model_sncosmo.set(x0=p['x0'], x1=p['x1'], c=p['c'])
            print(p)
            pars['X0'].full += p['x0']
            pars['X1'].full += p['x1'] + 0.13
            pars['c'].full += p['c']
            
            #model_sim.renorm()
            
            ms = sncosmo.get_magsystem('ab')
            bd = sncosmo.get_bandpass(band)
            
            mm = model_sim(pars)
            nn = model_sncosmo.bandflux(band, t+dt, zp=zp, zpsys='ab')[20]/mm[20]
            print(nn) 
            
            pl.figure()
            pl.plot(t, model_sncosmo.bandflux(band, t+dt, zp=zp, zpsys='ab'), label='sncosmo')
            pl.plot(t, mm*nn, label = 'NaCl * norm')
            pl.xlabel('phase')
            pl.ylabel('flux')
            pl.title(band+', X0=1, X1=c=0  norm ' + str(nn))
            pl.legend()
            pl.show()
            
            pl.figure()
            pl.plot(t, np.abs(mm[:len(t)]*nn-model_sncosmo.bandflux(band, t+dt, zp=zp, zpsys='ab'))/model_sncosmo.bandflux(band, t+dt, zp=zp, zpsys='ab'), label='relative difference')
            pl.xlabel('phase')
            pl.ylabel('flux')
            pl.title(band+', relative difference')
            pl.legend()
            pl.show()
    return tds_sim
        
    #pl.figure()
    #pl.plot(t, np.abs((model_sncosmo.bandflux('standard::b', t) - mm[:len(t)]*nn)/model_sncosmo.bandflux('standard::b', t)), label='sncosmo')
    #pl.xlabel('phase')
    #pl.ylabel('flux diff %')
    #pl.title('Diff Standard B, X0=1, X1=c=0')
    #pl.legend()
    #pl.show()
    
    #return model_sim



import logging
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s',
                    level=logging.INFO)


import os
import os.path as op
import numpy as np
import scipy.sparse
import sksparse.cholmod

import croaks
import saunerie
from saunerie.fitparameters import FitParameters

from salt3.minimizers import minimize_nr_var_quad, minimize_nr
import salt3.generator
import salt3.regularizations
import salt3.constraints
import salt3.models
import salt3.plots

from helpers import locate_test_datafile
from helpers import check_grad



def fit_1dsplinemodel_same_lcs(N=200, minimize=minimize_nr_var_quad, mu_regularization=500., plot=False, seed=None):
    """
    """
    lc_data = salt3.generator.generate_lc(N, tmax_range=(0., 0.),
                                          sig_range=(0., 0.),
                                          norm_range=(1000., 1000.),
                                          seed=seed)

    model = salt3.models.SplineModel1D(lc_data, grid_size=250)
    f = salt3.models.VarModelResiduals(lc_data, model)
    reg = salt3.constraints.SplineParamsRegularization(f.pars, mu=mu_regularization)
    cons = salt3.regularizations.NonIntegralConstraints(model.pars, model.spline_basis)
    
    # initialization fit (spline parameters only)
    f.pars['flux'].full[:] = 1000
    f.pars['theta'].full[:] = 1.E-6
    f.pars['flux'].fix()
    f.pars['tmax'].fix()
    f.pars['s'].fix()
    x = minimize(f, f.pars.free, reg=reg, cons=cons)
    x = x[0]
    f.pars.free = x

    # real fit, with all parameters free
    f.pars.release()
    f.pars['flux'].full[:] = 10000
    x = minimize(f, f.pars.free, cons=cons, reg=reg)
    f.pars.free = x
    
    if plot:
        import pylab as pl
        import matplotlib as mpl
        pl.figure()
        gs = mpl.gridspec.GridSpec(4,1)
        ax = pl.subplot(gs[0:3,0])
        pl.errorbar(lc_data.date, lc_data.flux, yerr=lc_data.flux_err, ls='', marker='.', color='b')
        xx = np.linspace(lc_data.date.min(), lc_data.date.max(), 100)
        xx.sort()

        try:
            tmax = f.pars['tmax'].full
        except:
            tmax = np.zeros(N)
        xdata = salt3.plots.get_xdata_for_model_evaluation(N, tmax, lc_data.band_set, npoints=200)
        v = model(f.pars.free, dp=xdata, jac=False)
        for sn in lc_data.sn_set:
            idx = xdata.sn == sn
            pl.plot(xdata.date[idx], v[idx], ls='-', color=pl.cm.jet(sn/xdata.sn.max()))
        
        pl.subplot(gs[3,0], sharex=ax)
        pl.errorbar(lc_data.date, f(x,jac=False)*lc_data.flux_err, yerr=lc_data.flux_err, marker='.', ls='', color='b')
        pl.grid(1)
        pl.xlabel('x')
        pl.ylabel('res')
        pl.subplots_adjust(hspace=0.05)        

        pl.figure()
        xx = np.linspace(-50., 50., 10000)
        jacobian = model.spline_basis.eval(xx)
        pl.plot(xx, jacobian @ model.pars['theta'].full, 'r-')
        pl.xlabel('x')
        pl.ylabel('LC template (spline model only)')
        pl.grid(1)
        
    return lc_data, f, cons, reg, x


def test_model_derivatives(N=200, seed=0):
    """
    """
    lc_data = salt3.generator.generate_lc(N=N, tmax_range=(-70., 70.),
                                          sig_range=(-1., 1.),
                                          norm_range=(10., 1000.),
                                          seed=seed)

    model = salt3.models.SplineModel1D(lc_data, grid_size=250)
    f = salt3.models.VarModelResiduals(lc_data, model)
    
    jacobian, JJ = check_grad(model, model.pars.free)
    D = np.abs(jacobian-JJ)
    np.testing.assert_allclose(np.zeros_like(D), D, atol=1.E-8)
    
    jacobian, JJ = check_grad(f, model.pars.free)
    D = np.abs(jacobian-JJ)
    np.testing.assert_allclose(np.zeros_like(D), D, atol=1.E-6)

    model.pars.free[:] += np.random.uniform(-1., 1., len(model.pars.free))
    jacobian, JJ = check_grad(model, model.pars.free)
    D = np.abs(jacobian-JJ)
    np.testing.assert_allclose(np.zeros_like(D), D, atol=1.E-6)
    
    jacobian,JJ = check_grad(f, model.pars.free)
    D = np.abs(jacobian-JJ)
    np.testing.assert_allclose(np.zeros_like(D), D, atol=1.E-6)

    # en revanche, celui la, il plante ... 
    #    model.pars.full[:] += np.random.uniform(-1., 1., len(model.pars.full))
    #    jacobian, JJ = check_grad(model, model.pars.free)
    #    D = np.abs(jacobian-JJ)
    #    np.testing.assert_allclose(np.zeros_like(D), D, atol=1.E-6)
    
    #    jacobian,JJ = check_grad(f, model.pars.free)
    #    D = np.abs(jacobian-JJ)
    #    np.testing.assert_allclose(np.zeros_like(D), D, atol=1.E-6)



    
def test_1dsplinemodelfullfit_nr():
    x_target = np.load(locate_test_datafile('test_1dsplinemodel_solution.npy'))
    lc, f, cons, reg, x = fit_1dsplinemodel_same_lcs(N=1000, minimize=minimize_nr, mu_regularization=500., plot=0, seed=0)
    np.testing.assert_allclose(x_target, x, atol=1.E-6)
    np.testing.assert_allclose(np.zeros_like(cons(x, jac=False)), cons(x, jac=False), atol=1.E-8)    
    


def fit_1dsplinemodel(N=200, minimize=minimize_nr, mu_regularization=500., plot=False, seed=None):
    """
    given
    """
    logging.info('generating data...')
    lc = salt3.generator.generate_lc(N=N, tmax_range=(-50., 50.),
                                     sig_range=(-0.5, 0.5),
                                     norm_range=(100., 10000.),
                                     seed=seed)
    model = salt3.models.SplineModel1D(lc, grid_size=250)

    # fit initialization
    logging.info('initialization...')    
    model.pars.full[:] = salt3.models.init_splinemodel1d(model).full[:]
    p_init = model.pars.copy()

    # minimization
    logging.info('minimization...')    
    reg = salt3.models.SplineParamsRegularization(model.pars, mu=mu_regularization)
    cons = salt3.models.NonIntegralConstraints(model.pars, model.spline_basis)
    f = salt3.models.VarModelResiduals(lc, model)
    x = minimize_nr(f, f.pars.free, cons=cons, reg=reg)

    if plot:
        import pylab as pl
        import matplotlib as mpl
        salt3.plots.plot_all_lcs_nrl(lc, model)
        salt3.plots.plot_all_lcs(lc, model)

        # plot the template itself 
        pl.figure()
        xx = np.linspace(-50., 50., 10000)
        jacobian = model.spline_basis.eval(xx)
        pl.plot(xx, jacobian @ model.pars['theta'].full, 'r-')
        pl.xlabel('x')
        pl.ylabel('LC template (spline model only)')
        pl.grid(1)
        pl.title('spline model')

        # and the residuals 
        pl.figure()
        pl.hist(f(f.pars.free, jac=False), bins=100)
        pl.xlabel('residuals')
    
    return lc, f, model, cons, reg, p_init, x
    

def test_1dsplinemodelfullfit_with_init_200():
    ff = np.load(locate_test_datafile('test_1dsplinemodel_fullfit_with_init_solution.npz'))
    x_target = ff['x_200']
    lc, f, model, cons, reg, p_init, x = fit_1dsplinemodel(N=200, minimize=minimize_nr, mu_regularization=200., plot=False, seed=0)
    np.testing.assert_allclose(x_target, x, atol=1.E-8)
    c = cons(x, jac=False)
    np.testing.assert_allclose(np.zeros_like(c), c, atol=1.E-8)

def test_1dsplinemodelfullfit_with_init_500():
    ff = np.load(locate_test_datafile('test_1dsplinemodel_fullfit_with_init_solution.npz'))
    x_target = ff['x_500']
    lc, f, model, cons, reg, p_init, x = fit_1dsplinemodel(N=500, minimize=minimize_nr, mu_regularization=500., plot=False, seed=0)
    np.testing.assert_allclose(x_target, x, atol=1.E-8)
    c = cons(x, jac=False)
    np.testing.assert_allclose(np.zeros_like(c), c, atol=1.E-8)
    
def test_1dsplinemodelfullfit_with_init_1000():
    ff = np.load(locate_test_datafile('test_1dsplinemodel_fullfit_with_init_solution.npz'))
    x_target = ff['x_1000']
    lc, f, model, cons, reg, p_init, x = fit_1dsplinemodel(N=1000, minimize=minimize_nr, mu_regularization=1000., plot=False, seed=0) 
    np.testing.assert_allclose(x_target, x, atol=1.E-8)
    c = cons(x, jac=False)
    np.testing.assert_allclose(np.zeros_like(c), c, atol=1.E-8)
    

    
if __name__ == "__main__":
    lc, f, cons, reg, x = fit_1dsplinemodel_same_lcs(N=1000, minimize=minimize_nr, mu_regularization=500., plot=True)
    lc, f, model, cons, reg, p_init, x = fit_1dsplinemodel(N=500, minimize=minimize_nr, mu_regularization=500., plot=True, seed=0)
    lc, f, model, cons, reg, p_init, x = fit_1dsplinemodel(N=2500, minimize=minimize_nr, mu_regularization=2500., plot=True, seed=0)    



#!/usr/bin/env python

import os
import sys
import time
from pathlib import Path
import logging
import pandas

import numpy as np
import pylab as pl

from sparse_dot_mkl import dot_product_mkl

from lemaitre import bandpasses
from nacl import TrainingDataset, LogLikelihood, Minimizer
from nacl.models import salt2
from nacl.models.salt2 import constraints
from nacl.models.salt2.regularizations import NaClSplineRegularization



def clean_dataset(tds, phase_range=(-25., +80.)):
    """
    """
    # clean ZTF19adcecwu
    idx = (tds.lc_data.sn == 'ZTF19adcecwu') & (tds.lc_data.mjd > 58840) & (tds.lc_data.flux<10000.)
    logging.info(f'removing {idx.sum()} outliers identified on ZTF19adcecwu g- and r- DR2 lightcurves')
    tds.lc_data.valid[idx] = 0

    # phase range (photometric data)
    tmax = np.zeros(len(tds.sn_data))
    tmax[tds.sn_data.sn_index] = tds.sn_data.tmax
    phot_tmax = tmax[tds.lc_data.sn_index]
    phase = (tds.lc_data.mjd - phot_tmax) / (1. + tds.lc_data.z)
    print(phase)
    print(phase_range)
    idx = (phase<phase_range[0]) | (phase>phase_range[1])
    logging.info(f'removing {idx.sum()} photometric points outside phase range')
    tds.lc_data.valid[idx] = 0

    # phase range (spectra)
    spec_tmax = tmax[tds.spec_data.sn_index]
    phase = (tds.spec_data.mjd - spec_tmax) / (1. + tds.spec_data.z)
    idx = (phase<phase_range[0]) | (phase>phase_range[1])
    logging.info(f'removing {idx.sum()} spectroscopic points outside phase range')
    tds.spec_data.valid[idx] = 0

    # points of the edge of the wavelength basis
    i_basis_max = tds.spec_data.i_basis.max()
    idx = (tds.spec_data.i_basis < 2) | (tds.spec_data.i_basis >= (i_basis_max-2))
    logging.info(f'removing {idx.sum()} spectroscopic points outside wavelength range')
    tds.spec_data.valid[idx] = 0

    # clean all SNe below 0.02
    #idx = (tds.lc_data.z < 0.01)
    #logging.info(f'removing {idx.sum()} very low redshift SNe')
    #tds.lc_data.valid[idx] = 0

    tds.compress()




def main(lc_only=False, spec_only=False,
         model_phase_range=(-20., +50.),
         data_phase_range=(-20., +40.)):
    # trainin dataset
    fl = bandpasses.get_filterlib()
    tds = TrainingDataset.read_parquet('lemaitre.compressed.parquet',
                                       filterlib=fl)
    clean_dataset(tds, phase_range=data_phase_range)

    if lc_only:
        tds = TrainingDataset(sne=tds.sn_data.nt,
                              lc_data=tds.lc_data.nt,
                              basis=tds.basis,
                              filterlib=fl)
    elif spec_only:
        tds = TrainingDataset(sne=tds.sn_data.nt,
                              spec_data=tds.spec_data.nt,
                              basis=tds.basis,
                              filterlib=fl)

    # model 
    model = salt2.SALT2Like(tds,
                            # phase_range=model_phase_range,
                            wl_grid=tds.basis.grid,
                            spectrum_recal_degree=3)
    
    # likelihood 
    ll = LogLikelihood(model)
    pars = model.init_pars()


    # minimizer
    mn = Minimizer(ll)
    
    # preliminary control plots : how does the model compare with the data ?
    v = model(pars)

    # recalibrate the spectra
    # r = recalib_spectra(tds, v)
    # if not lc_only:
    #    ll.pars['SpectrumRecalibration'].full[3::4] = np.log(r)
    #    pars['SpectrumRecalibration'].full[3::4] = np.log(r)

    return fl, tds, model, ll, pars, mn, v



if __name__ == '__main__':

    fl, tds, model, ll, pars, mn, v = main(lc_only=False)
    v, J = model(ll.pars, jac=1)
    J = J.tocsr()

    # use mkl
    t_start = time.perf_counter()
    H = dot_product_mkl(J.T, J, reorder_output=True)
    t_end = time.perf_counter()
    print(f'MKL: {t_end-t_start}, {H.shape}')




"""
"""

import sys
import logging

import numpy as np
from scipy.interpolate import interp1d
from scipy import sparse
import matplotlib.pyplot as plt

from sksparse import cholmod

from saltworks.plottools import binplot
from saltworks.robuststat import mad
from bbf.bspline import integ


class Spec:

    class FitResults:
        pass

    def __init__(self, tds, spec, basis,
                 error_eval_bin_width=50.,
                 beta=1.E-8):
        self.tds = tds
        self.spec = spec
        self.basis = basis
        self.error_eval_bin_width = error_eval_bin_width
        self.beta = beta
        if tds.spec_data != None:
            tds_spec_data = tds.spec_data
        else:
            tds_spec_data = tds.spectrophotometric_data
        self.tds_spec_data = tds_spec_data
        idx = self.idx = tds_spec_data.spec == spec

        for field in ['sn', 'mjd', 'valid', 'spec', 'exptime', 'z']:
            setattr(self, field, self._check_field(field))
        sn = tds_spec_data.sn[idx]
        assert np.all(np.equal(sn, sn[0]))

        self.wl = tds_spec_data.wavelength[idx]
        self.restframe_wl = self.wl / (1. + self.z)
        self.flux = tds_spec_data.flux[idx]
        self.fluxerr = tds_spec_data.fluxerr[idx]

        # # cut the NaN's
        # self.cut = np.isnan(self.flux) | (self.fluxerr < 0) | np.isnan(self.fluxerr)
        # if self.cut.sum() > 0:
        #     logging.warning(f'{self.cut.sum()} measurement detected with negative of nan uncertainties')

        # # cut the NaN's
        # self.cut = np.isnan(self.flux) | (self.fluxerr < 0) | np.isnan(self.fluxerr)
        # if self.cut.sum() > 0:
        #     logging.warning(f'{self.cut.sum()} measurement detected with negative of nan uncertainties')

        # wl_min, wl_max = basis.grid.min(), basis.grid.max()
        # # print(wl_min, wl_max, self.restframe_wl.min(), self.restframe_wl.max())
        # cut = (self.restframe_wl<wl_min) | (self.restframe_wl>wl_max)
        # if cut.sum() > 0:
        #     logging.info(f'{self.cut.sum()} outside basis range')
        # self.cut &= cut

        self.cut = self._select_data()

        self.fitres = []

    def _select_data(self):
        """the spectral data generally needs to be cleaned.
        """
        # remove the NaN's
        nan_idx = np.isnan(self.flux) | np.isnan(self.fluxerr)

        # some errors are negative
        # we don't cut the data - we resset them temporarily to 1.
        negflxerr_idx = self.fluxerr < 0.
        self.fluxerr[negflxerr_idx] = 1.

        # some data points are zeros
        zero_idx = (self.flux == 0.) | (self.fluxerr == 0.)

        # finally, get rid of all the data that is outside
        # the (restframe) basis wavelength rage
        wl_min, wl_max = self.basis.grid.min(), self.basis.grid.max()
        out_of_range_idx = (self.restframe_wl < wl_min) | (self.restframe_wl > wl_max)

        self.select_stats = {'nan': nan_idx.sum(),
                             'negative_errs': negflxerr_idx.sum(),
                             'zero_flux_or_err': zero_idx.sum(),
                             'out_of_range': out_of_range_idx.sum()}

        cut = nan_idx | zero_idx | out_of_range_idx
        logging.info(f'{self.sn}: {cut.sum()} measurement removed.')
        if cut.sum() > 0:
            logging.info(f'nan:{nan_idx.sum()} zflx:{zero_idx.sum()} oorng: {out_of_range_idx.sum()}')

        return cut

    def _check_field(self, name):
        """
        """
        try:
            s = np.unique(self.tds_spec_data.nt[self.idx][name])
        except:
            s = np.unique(self.tds_spec_data.__dict__[name][self.idx])
        assert len(s) == 1
        return s[0]

    def fit(self, x, y, yerr, beta=None):
        """
        """
        N = len(x)
        assert (len(y) == N) and (len(yerr) == N)
        if beta is None:
            beta = self.beta

        J = self.basis.eval(x)
        w = 1. / yerr
        W = sparse.dia_matrix((w**2, 0), shape=(N,N))
        H = J.T @ W @ J
        fact = cholmod.cholesky(H.tocsc(), beta=beta)

        r = Spec.FitResults()
        r.coeffs = fact(J.T @ W @ y)
        r.res = (y - J @ r.coeffs)
        r.wres = (y - J @ r.coeffs) * w
        r.chi2 = (r.wres**2).sum()
        r.ndof = (len(y) - len(r.coeffs))
        r.rchi2 = r.chi2 / r.ndof
        r.basis = self.basis
        #try:
        HH = H.todense() + np.diag(np.full(len(r.coeffs), 1.E-20))
        r.coeffs_cov = np.linalg.inv(HH)
        r.coeffs_err = np.sqrt(np.array(r.coeffs_cov.diagonal()).squeeze())
        r.i = np.arange(len(r.coeffs))
        r.selection = r.coeffs_err < 0.5/beta

            # r.coeffs_err = np.sqrt(scipy.sparse.linalg.inv(H).diagonal())
        #except:
        #    r.coeffs_err = np.zeros_like(r.coeffs)

        return r

    def recompute_error_model(self, wl, res, nbins=10):
        """
        """
        #       nbins = int((wl.max() - wl.min()) / bin_width)
        x, y, yerr = binplot(wl, res, nbins=nbins, scale=False, noplot=True)
        self.error_model = interp1d(x, yerr, kind='cubic', fill_value=(yerr[0], yerr[-1]), bounds_error=False)
        self.x_err, self.y_err = x, yerr
        return self.error_model

    def flag_residuals(self):
        """
        """
        pass

    def process(self):
        """
        """
        rwl = self.restframe_wl[~self.cut]
        flx = self.flux[~self.cut]
        flxerr = self.fluxerr[~self.cut]

        if len(rwl) == 0:
            logging.error(f'{self.sn} no data')
            return None

        # re-eval errors
        r = self.fit(rwl, flx, flxerr)
        error_model = self.recompute_error_model(rwl, r.res, nbins=10)
        self.fitres.append(r)

        # re-fit with recomputed errors
        model_flxerr = error_model(self.restframe_wl)
        if np.any(model_flxerr <= 0.):
            logging.error(f'{self.sn}: unable to recompute an error model')
            self.estimated_fluxerr = model_flxerr
            return None

        r = self.fit(rwl, flx, model_flxerr[~self.cut])
        self.fitres.append(r)

        # # bin spectrum
        # r = self.fit(self.wl, self.flux, fluxerr, order=1, bin_width=self.bin_width)
        # self.fitres.append(r)

        self.estimated_fluxerr = model_flxerr

    def get_projected_spectrum(self):
        """
        """
        if len(self.fitres) != 2:
            return None

        r = self.fitres[1]
        idx = r.selection
        wl = integ(r.basis, n=1) / integ(r.basis, n=0)
        # wl = wl[r.selection]
        N = idx.sum()

        d = np.zeros(N, self.tds_spec_data.nt.dtype)
        d['sn'] = self.sn
        d['spec'] = self.spec
        d['mjd'] = self.mjd
        d['valid'] = self.valid
        d['exptime'] = self.exptime
        d['wavelength'] = wl[idx]
        d['i_basis'] = r.i[idx]
        d['flux'] = r.coeffs[idx]
        d['fluxerr'] = r.coeffs_err[idx]
        return d

    def plot(self):
        """
        """
        try:
            r = self.fitres[1]
        except:
            logging.warning('problem with the fit: showing initial fit')
            try:
                r = self.fitres[0]
            except:
                r = None
                logging.warning('problem with the fit: no fit available')

        fig, axes = plt.subplots(nrows=4, ncols=1, figsize=(8,8), sharex=False)

        fig.suptitle(f'SN{self.sn} z={self.z} mjd={self.mjd}')

        # the original spectrum (with re-estimated errors and the fit)
        axes[0].errorbar(self.restframe_wl[~self.cut], self.flux[~self.cut], yerr=self.estimated_fluxerr[~self.cut],
                         ls='', marker='.', label='orig')
        wl_min, wl_max = self.restframe_wl.min(), self.restframe_wl.max()

        if r is not None:
            wl = np.linspace(wl_min, wl_max, 1000)
            J = self.basis.eval(wl)
            axes[0].plot(wl, J @ r.coeffs, 'r-', zorder=1000)
            axes[0].set_ylabel('spectrum')

        # the residuals
        if r is not None:
            axes[1].errorbar(self.restframe_wl[~self.cut], r.res, yerr=self.estimated_fluxerr[~self.cut],
                             ls='', marker='.', color='b')
            axes[1].set_ylabel('residuals')

            axes[2].plot(self.x_err, self.y_err, 'bo')
            xx = np.linspace(self.restframe_wl.min(), self.restframe_wl.max(), 100)
            axes[2].plot(xx, self.error_model(xx), 'r-')
            axes[2].set_ylabel('error model')

        # axes[3].shared_x_axes.remove(axes[3])
        if r is not None:
            axes[3].plot(r.i, r.coeffs,
                         color='gray', marker='.', alpha=0.5, ls='')
            axes[3].errorbar(r.i[r.selection], r.coeffs[r.selection], yerr=r.coeffs_err[r.selection],
                             ls='', marker='.')
            y = r.coeffs[r.selection]
            ym, ys = np.median(y), mad(y)
            axes[3].set_ylim((ym-3*ys, ym+3*ys))
            axes[3].set_ylabel('projection')


def clean_and_project_spectra(tds, basis):
    """
    """
    #spec_data = tds.spec_data
    if tds.spec_data is not None:
        spec_data = tds.spec_data
    else:
        spec_data = tds.spectrophotometric_data
    l = []
    with_errors = []
    z = spec_data.z
    for spec in spec_data.spec_set:
        s = Spec(tds, spec, basis)
        logging.info(f'processing {s.sn} {spec}')
        try:
            s.process()
            p = s.get_projected_spectrum()
        except:
            logging.error(f'unable to process: {s.sn}')
            logging.error(f'{sys.exc_info()}')
            p = None
        if p is None:
            with_errors.append(s)
            continue
        l.append(p)
    return l, with_errors



#!/usr/bin/env python3

import os
import os.path as op
import sys

import numpy as np
import pandas

from lemaitre import bandpasses
from nacl import TrainingDataset
from nacl.train import ModelTrainer
from nacl.models.salt2 import SALT2Like



if __name__ == '__main__':
    """
    """
    fl = bandpasses.get_filterlib()
    tds = TrainingDataset.read_parquet('lml.compressed.parquet', filterlib=fl)
    model = SALT2Like(tds)
    train = ModelTrainer(tds, model)

    # sne_ztf = pandas.read_parquet('data_ztf_blind.parquet.gzip')
    # sne_ztf['col'] = sne_ztf['c']
    # lc_ztf = pandas.read_parquet('lc_SNLS_blind.parquet.gzip')
    # lc_ztf['x'] = 0.
    # lc_ztf['y'] = 0.
    # lc_ztf['sensor_id'] = 12
    # sp_ztf = pandas.read_parquet('sp_SNLS_blind.parquet.gzip')
    # sp_ztf['i_basis'] = 0

    # sne_snls = pandas.read_parquet('data_SNLS_blind.parquet.gzip')
    # sne_snls['col'] = sne_snls['c']
    # lc_snls = pandas.read_parquet('lc_SNLS_blind.parquet.gzip')
    # lc_snls['x'] = 0.
    # lc_snls['y'] = 0.
    # lc_snls['sensor_id'] = 12
    # sp_snls = pandas.read_parquet('sp_SNLS_blind.parquet.gzip')
    # sp_snls['i_basis'] = 0

    # lc_snls.lc += lc_ztf.lc.max() + 1
    # sp_snls.spec += sp_ztf.spec.max() + 1

    # tds = TrainingDataset(pandas.concat([sne_ztf, sne_snls]),
    #                       pandas.concat([lc_ztf, lc_snls]),
    #                       pandas.concat([sp_ztf, sp_snls]),
    #                       filterlib=fl)



import nacl.simulations.ideal as ideal
from nacl.models.salt import SALT2Like
from nacl.simulations.flux_simulator import Simulator
from nacl.models.variancemodels import SimpleVarianceModel
from nacl.minimizers import Minimizer
from nacl.fit import FitModel2D
import numpy as np 

N_SN = 500
MU_REG = 10

####################################### SIMULATION #############################
# Create the cadence
DS = ideal.Generator(n_sn=N_SN, lsst_simu=True)

# Simulate the flux and the measurements error
INIT_FROM_SALT2_FILE = "../data/salt2.npz"
FS = Simulator(DS,                                         # DataSet 
               SALT2Like,                                  # Model function used to simulated
               variance_model_func=SimpleVarianceModel,    # Error Snake used in sim
               gamma_init=0,                               # Value of the error snake
               init_from_salt2_file=INIT_FROM_SALT2_FILE)  # Model values for sim

####################################### FIT ####################################
## Instantiate Fit
# Create Constraints and Regularization
# as function of free parameters.
fit = FitModel2D(FS.trainingDataset,                           # TrainingDataset
                 model_func=SALT2Like,                         # Model used
                 normalization_band_name='SWOPE::B',           # Model normalization
                 init_from_salt2_file=INIT_FROM_SALT2_FILE,    # Files to initiate the model
                 pars_fix=[],                                  # Parameters that should be fixed
                 mu_reg=MU_REG,                                # Regularization hyperparameter 
                 minimizer=Minimizer)                          # Minimizer function

# degrade SN parameters (multiply by N(1, 0.05))
# degrade error snake  (multiply by N(1, 1))
fit.parameter_degradation(model_coef_init=0.05,
                          var_coef_init=0.1)

# Fit model parameters
fit(pars_fix=['eta_calib', 'kappa_color', 'gamma'])





# Projects the flux on a 1D BSpline of order 0
# The idea is to simplify the evaluation of the model in a first place

import logging
from multiprocessing import Pool

import scipy.sparse as sparse
from sksparse import cholmod
import numpy as np
import scipy
import scipy.sparse.linalg
from scipy.interpolate import interp1d
import pylab as pl
import pandas as pd

from bbf.bspline import BSpline, BSpline2D
# from nacl.models import import salt2
from nacl.dataset import TrainingDataset
from saunerie.plottools import binplot


class Spec:

    class FitResults:
        pass

    def __init__(self, tds, spec, basis=None, error_eval_bin_width=50.,
                 bin_width=40., beta=1.E-8):
        self.tds = tds
        self.spec = spec
        self.basis = basis
        self.error_eval_bin_width = error_eval_bin_width
        self.bin_width = bin_width
        self.beta = beta

        idx = tds.spec_data.spec == spec
        self.wl = tds.spec_data.wavelength[idx]
        self.flux = tds.spec_data.flux[idx]
        self.fluxerr = tds.spec_data.fluxerr[idx]
        self.cut = (self.fluxerr < 0) | np.isnan(self.fluxerr)
        if self.cut.sum() > 0:
            logging.warning(f'{self.cut.sum()} measurement detected with negative of nan uncertainties')
        sn = tds.spec_data.sn[idx]
        assert np.all(np.equal(sn, sn[0]))
        self.idx = idx

        for field in ['sn', 'name', 'mjd', 'valid', 'spec', 'exptime']:
            setattr(self, field, self._check_field(field))

        self.fitres = []

    def _check_field(self, name):
        s = np.unique(self.tds.spec_data.nt[self.idx][name])
        assert len(s) == 1
        return s[0]

    def fit(self, x, y, yerr, bin_width, order=1, beta=None):
        """
        """
        N = len(x)
        assert (len(y) == N) and (len(yerr) == N)
        if beta is None:
            beta = self.beta

        if self.basis is None:
            grid = np.arange(x.min(), x.max()+bin_width, bin_width)
            basis = BSpline(grid, order=order)
        else:
            basis = self.basis

        J = basis.eval(x)
        w = 1. / yerr
        W = sparse.dia_matrix((w**2, 0), shape=(N,N))
        H = J.T @ W @ J
        fact = cholmod.cholesky(H.tocsc(), beta=beta)

        r = Spec.FitResults()
        r.coeffs = fact(J.T @ W @ y)
        r.res = (y - J @ r.coeffs)
        r.wres = (y - J @ r.coeffs) * w
        r.chi2 = (r.wres**2).sum()
        r.ndof = (len(y) - len(r.coeffs))
        r.rchi2 = r.chi2 / r.ndof
        r.basis = basis
        #try:
        HH = H.todense() + np.diag(np.full(len(r.coeffs), 1.E-20))
        r.coeffs_cov = np.linalg.inv(HH)
        r.coeffs_err = np.sqrt(np.array(r.coeffs_cov.diagonal()).squeeze())

            # r.coeffs_err = np.sqrt(scipy.sparse.linalg.inv(H).diagonal())
        #except:
        #    r.coeffs_err = np.zeros_like(r.coeffs)

        return r

    def recompute_error_model(self, wl, res, nbins=10):
        """
        """
        #       nbins = int((wl.max() - wl.min()) / bin_width)
        x, y, yerr = binplot(wl, res, nbins=nbins, scale=False, noplot=True)
        self.error_model = interp1d(x, yerr, kind='cubic', fill_value=(yerr[0], yerr[-1]), bounds_error=False)
        self.x_err, self.y_err = x, yerr
        return self.error_model

    def process(self):
        """
        """
        # re-eval errors
        r = self.fit(self.wl, self.flux, self.fluxerr, order=4, bin_width=self.error_eval_bin_width)
        error_model = self.recompute_error_model(self.wl, r.res, nbins=10)
        self.fitres.append(r)

        # re-fit with recomputed errors
        fluxerr = error_model(self.wl)
        r = self.fit(self.wl, self.flux, fluxerr, order=4, bin_width=self.error_eval_bin_width)
        self.fitres.append(r)

        # bin spectrum
        r = self.fit(self.wl, self.flux, fluxerr, order=1, bin_width=self.bin_width)
        self.fitres.append(r)

        self.estimated_fluxerr = fluxerr

    def get_binned_spectrum(self):
        r = self.fitres[2]
        N = len(r.coeffs)
        d = np.zeros(N, self.tds.spec_data.nt.dtype)
        d['sn'] = self.sn
        d['name'] = self.name
        d['mjd'] = self.mjd
        d['valid'] = self.valid
        d['exptime'] = self.exptime
        d['wavelength'] = 0.5 * (r.basis._grid[1:] + r.basis._grid[:-1])
        d['flux'] = r.coeffs
        d['fluxerr'] = r.coeffs_err
        return d

    def plot(self):
        """
        """
        r = self.fitres[1]
        rr = self.fitres[2]

        fig, axes = pl.subplots(nrows=4, ncols=1, figsize=(8,8), sharex=True)

        # the original spectrum
        axes[0].errorbar(self.wl[~self.cut], self.flux[~self.cut], yerr=self.estimated_fluxerr[~self.cut],
                         ls='', marker='.', label='orig')

        grid_wl = 0.5 * (rr.basis._grid[1:] + rr.basis._grid[:-1])
        # axes[0].errorbar(grid_wl, rr.coeffs, yerr=rr.coeffs_err, marker='.', ls='', color='c')
        axes[0].plot(grid_wl, rr.coeffs, marker='.', ls='', color='c')
        #        axes[0].plot(grid_wl, coeffs, 'ro', label='binned', zorder=100)
        # axes[0].plot(wl, J @ coeffs, 'r-', label='binned', zorder=100)

        axes[1].errorbar(self.wl[~self.cut], r.res[~self.cut], yerr=self.estimated_fluxerr[~self.cut],
                         ls='', marker='.', color='b')

        axes[2].plot(self.x_err, self.y_err, 'bo')
        xx = np.linspace(self.wl.min(), self.wl.max(), 100)
        axes[2].plot(xx, self.error_model(xx), 'r-')

        # the binned spectrum
        axes[3].errorbar(grid_wl, rr.coeffs, yerr=rr.coeffs_err, marker='.', ls='', color='b')
        # axes[3].errorbar(self.wl, rr.res/self.error_model(self.wl), yerr=1., ls='', marker='.', color='b')


def rebin_spectra_2(tds):
    """
    """
    spec_data = tds.spec_data
    nb_spectra = len(tds.spec_data.spec_set)
    rebinned_spectra = []
    spectra_in_error = []

    def rebin_spectrum(spec):
        logging.info(f'processing {spec}')
        try:
            s = Spec(tds, spec, bin_width=40.)
            s.process()
            rebinned_spectra.append(s.get_binned_spectrum())
        except:
            logging.error(f'unable to process {spec}')
            spectra_in_error.append(s)

    pool = Pool()
    pool.map(rebin_spectrum, tds.spec_data.spec_set)

    all_spec = np.hstack(rebinned_spectra)
    return TrainingDataset(tds.sn_data, lc_data=tds.lc_data, spec_data=tds.spec_data)

def rebin_spectra(tds):
    """
    """
    spec_data = tds.spec_data
    nb_spectra = len(tds.spec_data.spec_set)
    c = 0
    rebinned_spectra = []
    spectra_in_error = []
    for spec in tds.spec_data.spec_set:
        logging.info(f'processing {spec} {c}/{nb_spectra}')
        try:
            s = Spec(tds, spec, bin_width=40.)
            s.process()
            rebinned_spectra.append(s.get_binned_spectrum())
        except:
            logging.error(f'unable to process {spec}')
            spectra_in_error.append(s)
            continue
        c += 1

    all_spec = np.rec.array(np.hstack(rebinned_spectra))
    return TrainingDataset(tds.sn_data.nt, lc_data=tds.lc_data.nt, spec_data=all_spec)
    # return rebinned_spectra


def project_on_nacl_basis(tds, model):
    """
    """
    spec_data = tds.spec_data
    basis = model.basis.bx

    for spec in tds.spec_data.spec_set:
        try:
            s = Spec(tds, spec, basis=basis, z=None)
            s.process()
            rebinned_spectra.append(s.get_binned_spectrum())
        except:
            logging.error(f'unable to process {spec}')
            spectra_in_error.append(s)
            continue




def bin_spectrum(tds, spec,
                 error_eval_bin_with=50.,
                 bin_with=40.,
                 plot=False, beta=1.E-8):
    """A smaller rewrite of Mahmoud's original code
    """
    #    for i in tds.spec_data.spec_set:
    idx = tds.spec_data.spec == spec
    wl = tds.spec_data.wavelength[idx]
    flx = tds.spec_data.flux[idx]
    flxerr = tds.spec_data.fluxerr[idx]
    cut = (flxerr < 0) | np.isnan(flxerr)
    if cut.sum() > 0:
        logging.warning(f'{cut.sum()} measurement detected with negative of nan uncertainties')

    sn = tds.spec_data.sn[idx]
    assert np.all(np.equal(sn, sn[0]))
    N = len(sn)

    grid = np.arange(wl.min(), wl.max()+step, step)
    basis = BSpline(grid, order=4)
    # coeffs = basis.linear_fit(wl, flx, beta=1.E-10)

    # re-estimate the error bars, using an order-4 spline
    grid = np.arange(wl.min(), wl.max()+error_eval_bin_with, error_eval_bin_with)
    basis = BSpline(grid, order=4)
    J = basis.eval(wl)
    w = 1. / flxerr
    W = sparse.dia_matrix((w**2, 0), shape=(N,N))
    H = J.T @ W @ J
    fact = cholmod.cholesky(H, beta=beta)
    coeffs = fact(J.T @ W @ flx)
    res = (flx - J @ coeffs)
    x, y, yerr = binplot(wl, res, nbins=10, scale=False, noplot=True)
    wres = (flx - J @ coeffs) * w
    chi2 = (wres**2).sum()
    rchi2 = chi2 / (len(flx) - len(coeffs))

    # refit and bin
    grid = np.arange(wl.min(), wl.max()+bin_with, bin_with)
    basis = BSpline(grid, order=0)
    J = basis.eval(wl)
    w = 1. / flxerr
    W = sparse.dia_matrix((w**2, 0), shape=(N,N))
    H = J.T @ W @ J
    fact = cholmod.cholesky(H, beta=beta)
    coeffs = fact(J.T @ W @ flx)
    res = (flx - J @ coeffs)
    wres = (flx - J @ coeffs) * w
    chi2 = (wres**2).sum()
    rchi2 = chi2 / (len(flx) - len(coeffs))




    if plot:
        fig, axes = pl.subplots(nrows=4, ncols=1, figsize=(8,8), sharex=True)
        axes[0].errorbar(wl[~cut], flx[~cut], yerr=self.estimated_fluxerr[~cut],
                         ls='', marker='.', label='orig')
        grid_wl = 0.5 * (grid[1:] + grid[:-1])
        #        axes[0].plot(grid_wl, coeffs, 'ro', label='binned', zorder=100)
        axes[0].plot(wl, J @ coeffs, 'r-', label='binned', zorder=100)

        axes[1].errorbar(wl[~cut], res[~cut], yerr=flxerr[~cut], ls='', marker='.', color='b')
        x,y,yerr = binplot(wl, res, nbins=10, scale=False, ax=axes[1])
        axes[2].plot(x, yerr, 'bo')
        # p = np.polyfit(x, yerr, deg=3)
        p = interp1d(x, yerr, kind='cubic', fill_value=(yerr[0], yerr[-1]), bounds_error=False)

        xx = np.linspace(x.min(), x.max(), 100)
        axes[2].plot(xx, p(xx), 'r-')

        axes[3].errorbar(wl, res/p(wl), yerr=1., ls='', marker='.', color='b')

    return grid, coeffs, basis, fact, H, res, wres, chi2, rchi2



def tds_spec_data_binning(tds, step = 40.0):
    sn_data = tds.sn_data
    if tds.lc_data is None:
        spec_data = tds.spectrophotometric_data
    else:
        spec_data = tds.spec_data
        lc_data = tds.lc_data
    spec_index = spec_data.spec_index
    spec_flux = spec_data.flux
    spec_fluxerr = spec_data.fluxerr
    spec_wl = spec_data.wavelength
    spec_mjd = spec_data.mjd
    spec_sn = spec_data.sn
    spec_spec = spec_data.spec
    spec_valid = spec_data.valid
    
    
    SN = np.array([])
    MJD = np.array([])
    WL = np.array([])
    FLX = np.array([])
    FLXERR = np.array([])
    VALID = np.array([])
    SPEC = np.array([])
    #EXPTIME = np.array([])
    
    spec_not_inv = np.array([])
    
    for i in range( int( np.max(spec_index) ) ):
    #for i in range(1):
        idx = (spec_index == i)
        
        flux = spec_flux[idx]
        fluxerr = spec_fluxerr[idx]
        wl = spec_wl[idx]
        mjd = spec_mjd[idx]
        sn = spec_sn[idx]
        valid = spec_valid[idx]
        spec = spec_spec[idx]
        
        grid = np.arange(np.min(wl), np.max(wl)+step, step)
        wl_theta = 0.5*(grid[1:] + grid[:-1])
        try:
            base = BSpline(grid, order=1)
            
            J = base.eval(wl)
            W=sparse.dia_matrix((1./fluxerr**2, 0), shape=(len(mjd), len(mjd)))
            H = J.T @ W @ J
            f = cholmod.cholesky(H, beta=1e-10)
            theta = f(J.T @ W @ flux)
            #print(len(theta))
            #flux_new = J @ theta

            H_inv = np.linalg.inv(H.todense())
            theta_err = np.sqrt( np.diag (H_inv) )
            
            SN = np.append(SN, np.array([ sn[0] for j in range( len(theta) ) ]))
            MJD = np.append(MJD, np.array([ mjd[0] for j in range( len(theta) ) ]))
            WL = np.append(WL, wl_theta)
            FLX = np.append(FLX, theta)
            FLXERR = np.append(FLXERR, theta_err)
            VALID = np.append(VALID, np.array([ valid[0] for j in range( len(theta) ) ]))
            SPEC = np.append(SPEC, np.array([ spec[0] for j in range( len(theta) ) ]))
        except: #manual binning when fit doesn't work due to issues
            spec_not_inv = np.append(spec_not_inv, i)
            ii = np.digitize(wl, grid)-1
            grid_augment = grid[ii]
            flx_ = np.array([])
            flxerr_ = np.array([])
            wl_ = np.array([])
            for j in range(len(grid)-1):
                wave = grid[j]
                idx2 = grid_augment == wave
                flx = flux[idx2]
                if len(flx)==0:
                    continue
                else:
                    flxerr = fluxerr[idx2]
                    flux_mean = np.sum(flx/flxerr**2)/np.sum(1/flxerr**2)
                    fluxerr_mean =  np.sqrt(1/np.sum(1/flxerr**2))
                    
                    flx_ = np.append(flx_, flux_mean)
                    flxerr_ = np.append(flxerr_, fluxerr_mean)
                    wl_ = np.append(wl_, 0.5*(grid[j] + grid[j+1]))
            SN = np.append(SN, np.array([ sn[0] for j in range( len(flx_) ) ]))
            MJD = np.append(MJD, np.array([ mjd[0] for j in range( len(flx_) ) ]))
            WL = np.append(WL, wl_)
            FLX = np.append(FLX, flx_)
            FLXERR = np.append(FLXERR, flxerr_)
            VALID = np.append(VALID, np.array([ valid[0] for j in range( len(flx_) ) ]))
            SPEC = np.append(SPEC, np.array([ spec[0] for j in range( len(flx_) ) ]))
                
        
    EXPTIME = np.array([ np.nan for j in range( len(SN) ) ])
    SPEC_DATA = pd.DataFrame({'sn':SN, 'mjd':MJD,  'wavelength':WL, 'flux':FLX, 'fluxerr':FLXERR, 'valid':VALID, 'spec':SPEC, 'exptime':EXPTIME})
    SN_DATA = pd.DataFrame({'sn':sn_data.sn, 'z':sn_data.z, 'tmax':sn_data.tmax, 'x1':sn_data.x1, 'x0':sn_data.x0, 'col':sn_data.col, 'valid':sn_data.valid})
    if tds.lc_data is None:
        tds_new = TrainingDataset(SN_DATA, spectrophot_data = SPEC_DATA)
    else:
        LC_DATA = pd.DataFrame({'sn':lc_data.sn, 'mjd':lc_data.mjd, 'flux':lc_data.flux, 'fluxerr':lc_data.fluxerr, 'band':lc_data.band, 'magsys':lc_data.magsys, 'zp':lc_data.zp, 'exptime':lc_data.exptime,'valid':lc_data.valid, 'lc':lc_data.lc, 'seeing':lc_data.seeing, 'mag_sky':lc_data.mag_sky})
        tds_new = TrainingDataset(SN_DATA, lc_data = LC_DATA, spec_data = SPEC_DATA)
    return tds_new



"""The salt2 model
"""

import logging
import numpy as np

from .constraints import salt2like_linear_constraints, salt2like_classical_constraints
from .regularizations import NaClSplineRegularization
from .variancemodels import SimpleErrorSnake, CalibErrorModel, SNLambdaErrorSnake, LocalErrorSnake, SNLocalErrorSnake
from .salt import SALT2Like
from .spectra import *


def wl_extend(n_wl = 200, min_wl = 2000., max_wl = 11000., step=30.):
    """
    Increases resolution in a defined region of wavelength.
    """
    #bad_zone = [(3500.,3800.),(5900.,6200.),(7700.,8300.)]
    reg_zone = (min_wl,max_wl)
    basis = np.linspace(reg_zone[0], reg_zone[1], n_wl)
    #for b in bad_zone:
    #    B = np.arange( b[0], b[1] + step, step )
    #    basis = np.append( basis, B )
    #    basis.sort()
    #    basis = np.unique(basis)
    return basis
    
    
def phase_extend(n_p = 20, min_p = -20., max_p = 50., step=1.):
    """
    Increases resolution in a defined region of wavelength.
    """
    #bad_zone = [(-20.,-15.),(40.,50.)]
    reg_zone = (min_p,max_p)
    basis = np.linspace(reg_zone[0], reg_zone[1], n_p)
    #for b in bad_zone:
    #    B = np.arange( b[0], b[1] + step, step )
    #    basis = np.append( basis, B )
    #    basis.sort()
    #    basis = np.unique(basis)
    return basis
phase_grid = phase_extend()
wl_grid = wl_extend(n_wl = 127, max_wl = 9000.)

def get_model(tds, init_from_salt2_file='salt2.npz',
              normalization_band_name='SWOPE2::B'):
    """a general function to instantiate the model
    """
    ret = SALT2Like(tds,
                    wl_grid = wl_grid,
                    phase_grid = phase_grid)
    return ret

def get_constraint_prior(model, linear=True, mu=1.E6, Mb=-19.5, check=True):
    """instantiate a set of constraints to apply to the model parameters
    """
    cons = None
    if linear:
        cons = salt2like_linear_constraints(model, mu=mu, Mb=Mb)
    else:
        cons = salt2like_classical_constraints(model, mu=mu, Mb=Mb)
    # check active
    # if none active: return None
    return cons


def get_regularization_prior(model, pars, mu=1., order=1, check=True):
    """instantiate a regularization class
    """
    to_regularize = []
    for block_name in ['M0', 'M1']:
        if ((pars[block_name].indexof() >= 0).any()):
            to_regularize.append(block_name)
        else:
            logging.info(f'{block_name} has all pars fixed: no regularization needed')

    if check and len(to_regularize) == 0:
        return None

    return NaClSplineRegularization(model,
                                    pars,
                                    to_regularize=to_regularize,
                                    order=order,
                                    mu=mu)

def get_simple_snake_error(model):
    """instantiates a simple error snake model
    """
    variance_model = SimpleErrorSnake(model)
    return variance_model

def get_local_snake_error(model):
    """instantiates a local error snake model
    """
    variance_model = LocalErrorSnake(model)
    return variance_model
  
def get_sn_lambda_snake_error(model):
    """
    instantiates an error snake which is SN and wavelength dependent
    """
    variance_model = SNLambdaErrorSnake(model)
    return variance_model
    
def get_sn_local_snake_error(model):
    """
    """
    variance_model = SNLocalErrorSnake(model)
    return variance_model
    



from .utils import ToyModelGenerator
from .model import SingleBandLightcurveModel1D, LightCurveModel1D, LightCurveModel1DWithCalibScatter, LightCurveModel1DWithCalibAndColorScatter
from .regul import Regularization
from .constraints import cons
from .variancemodels import SimplePedestalModel, SimpleErrorSnake
from .priors import CalibPrior, ColorScatterPrior




import scipy.sparse as sparse
import numpy as np
from saltworks.fitparameters import FitParameters


def check_grad(model, pars, dx=1.E-6):
    """
    """
    v, jacobian = model(pars, jac=True)
    pars = pars.copy()
    df = []
    for i in range(len(pars.full)):
        k = pars.indexof(i)
        if k < 0:
            continue
        pars.full[i] += dx
        vp = model(pars, jac=False)
        df.append((vp-v)/dx)
        pars.full[i] -= dx
    return np.array(jacobian.todense()), np.vstack(df).T


def check_deriv(pen, pars, dx=1.E-6):
    """
    """
    v, grad, hess = pen(pars.free, deriv=True)
    pp = pars.copy()

    df, d2f = [], []
    for i in range(len(pp.full)):
        k = pars.indexof(i)
        if k < 0:
            continue
        pp.full[i] += dx
        vp = pen(pp.free, deriv=False)
        pp.full[i] -= (2*dx)
        vm = pen(pp.free, deriv=False)
        df.append((vp-vm)/(2*dx))
        pp.full[i] += dx
    return np.array(grad), np.vstack(df).T


def check_deriv_old(pen, pars, dx=1.E-6):
    """Temporary version, to accomodate the LogLikelihood transitional interface
    """
    v, grad, hess = pen(pars.free, deriv=True)
    pp = pars.copy()

    df, d2f = [], []
    for i in range(len(pp.full)):
        pp.full[i] += dx
        vp = pen(pp.free, deriv=False)
        pp.full[i] -= (2*dx)
        vm = pen(pp.free, deriv=False)
        df.append((vp-vm)/(2*dx))
        pp.full[i] += dx
    return np.array(grad), np.vstack(df).T



import numpy as np
import scipy.sparse
import croaks
from saunerie.fitparameters import FitParameters
from salt3.minimizers import minimize_nr_var_quad, minimize_nr
import salt3.models


def gen_c1poly(N=100, a=-0.1, b=-0.2, c=1.2, d=-0.2, e=1.2, ey=0.1,
               seed=None):
    """
    Generate a test dataset
    """

    if isinstance(seed, int):
        np.random.seed(seed)
    
    x = np.random.uniform(-10, 10., N)
    y = np.zeros(N)
    idx = x<0.
    y[idx] = a*x[idx]**2 + b*x[idx] + c
    idx = x>=0.
    y[idx] = d*x[idx] + e
    y += np.random.normal(scale=ey, size=N)
    ret = np.rec.fromarrays((x,y,np.full(N, ey)), names=['x', 'y', 'ey'])
    ret = croaks.DataProxy(ret, flux = 'y', wavelength = 'x', flux_err = 'ey')
    return ret


class C1PolyResiduals():
    
    def __init__(self, data):
        self.data = data.nt
        self.pars = FitParameters([('p1', 3), ('p2', 2)])
        
    def __call__(self, p, jac=False):
        self.pars.free = p
        a,b,c = self.pars['p1'].full
        d,e = self.pars['p2'].full
        x,y,ey = self.data['x'], self.data['y'], self.data['ey']
        N = len(x)
        v = np.zeros(N)

        idx = x<0
        v[idx] = a*x[idx]**2 + b*x[idx] + c
        idx = x>=0
        v[idx] = d*x[idx] + e
        o = np.ones(N)
        
        if jac:
            i = np.tile(np.arange(N), 5)
            j = np.repeat(np.arange(5), N)
            d = np.vstack((x**2/ey, x/ey, o/ey, x/ey, o/ey)).T
            idx = d[:,1]<0
            ii = np.where(idx)[0]
            d[ii, 3] = d[ii, 4] = 0.
            idx = d[:,1]>=0
            ii = np.where(idx)[0]
            d[ii,0] = d[ii,1] = d[ii,2] = 0
            jacobian = scipy.sparse.coo_matrix((-d.T.flatten(), (i,j)), shape=(N,5))
        
            return (y-v)/ey, jacobian

        return (y-v)/ey

    
class C1PolyConstraints():
    
    def __init__(self, pars):
        """
        """
        self.pars = pars # .copy()
        self.pars_free = pars.copy()
        self.pars_free.release()
        n = len(self.pars.full)
        i = [2,   4,  1,   3]
        j = [0,   0,  1,   1]
        v = [1., -1., 1., -1.]
        self.H = scipy.sparse.coo_matrix((v, (i,j)), shape=(n, 2)).T
        self.d = self.vals = np.array([0., 0.])
        
    #    def __len__(self):
    #        return 2

    def get_rhs(self):
        return self.d
    
    def __call__(self, p, jac=False):
        """
        """
        self.pars.free = p
        self.pars_free.full[:] = self.pars.full[:]

        # the constraint class return the delta with the target value 
        v = -self.H @ self.pars.full + self.vals
        
        if jac:
            return v, self.H
        return v        

    
class C1PolyRegz():
    """
    This one is not constraint free
    """
    def __init__(self, pars, mu=1.E-6):
        self.pars = pars
        d = np.full(5, mu)
        self.P = scipy.sparse.dia_matrix((d, [0]), shape=(5,5))
        print(self.P.todense())
        
    def __call__(self, p):
        self.pars.free = p
        v = self.pars.full.T.dot(self.P.dot(self.pars.full))
        return v, self.P


def main(N=100, plot=False, minimizer=minimize_nr_var_quad, **kw):
    """
    """
    
    data = gen_c1poly(N=N, **kw)
    m = C1PolyResiduals(data)
    cons = C1PolyConstraints(m.pars)
    regz = C1PolyRegz(m.pars, mu=1.E-8)
    f = salt3.models.VarModelResiduals(data, m)
    x = minimizer(f, m.pars.free, cons = cons, reg = regz)    

    if plot:
        import pylab as pl
        data = data.nt
        pl.figure()
        pl.errorbar(data['x'], data['y'], yerr=data['ey'], ls='', marker='.', color='b')
        d = np.rec.fromarrays((np.linspace(-10., 10, 500),
                               np.zeros(500),
                               np.ones(500)), names=['x', 'y', 'ey'])
        d = croaks.DataProxy(d, flux = 'y', wavelength = 'x', flux_err = 'ey')
        mm = C1PolyResiduals(d)
        v = mm(x[0])
        pl.plot(d.nt['x'], -v, 'r-')
        pl.xlabel('x')
        pl.ylabel('y')
    if len(x) == 2:
        x = x[0]
    return data, m, cons, x
    

def test_c1polyfit_nr():
    dp, f, cons, x = main(minimizer=minimize_nr_var_quad,
                          plot=0, seed=0)
    np.testing.assert_allclose(np.array([-0.10092981, -0.2041651 ,  1.23490678, -0.2041651 ,  1.23490678]), x, atol=1.E-6)
    np.testing.assert_allclose(np.zeros_like(cons(x, jac=False)), cons(x, jac=False), atol=1.E-8)

    
def test_c1polyfit_nr_qr():
    dp, f, cons, x = main(minimizer=minimize_nr_var_quad, plot=0, seed=0)
    np.testing.assert_allclose(np.array([-0.10092981, -0.2041651 ,  1.23490678, -0.2041651 ,  1.23490678]), x, atol=1.E-6)
    np.testing.assert_allclose(np.zeros_like(cons(x, jac=False)), cons(x, jac=False), atol=1.E-8)

    
if __name__ == "__main__":
    dp, f, cons, x = main(minimizer=minimize_nr_var_quad, plot=1)



"""This module
"""


import codecs
import os
import base64
import urllib
import keyring
import socket

import numpy as np
import scipy.sparse as sparse


class Credentials:
    """A helper class to manage the HTTP/basic credentials
    """
    def __init__(self, username, **kwargs):
        self.username = username
        self._password = kwargs.get('password', None)
        self.service = kwargs.get('service', 'nacl')
        self.realm = kwargs.get('realm', None)

    def __repr__(self):
        l = f'Credentials: realm/user: {self.realm}/{self.username}'
        if self._password is None:
            passwd = keyring.get_password(self.service, self.username)
            if passwd is None:
                l += ' | passwd: unknown'
            else:
                l += ' | passwd: <from-keyring>'
        else:
            l += ' | passwd: ******'
        return l

    @property
    def password(self):
        if self._password is not None:
            return self._password
        if self.service is not None:
            passwd = keyring.get_password(self.service, self.username)
            if not passwd:
                raise ValueError(f'no password in keyring for ({self.service}/{self.username})')
            return passwd


def _download_file(remote_url, target, credentials=None):
    """
    Accepts a URL, downloads the file to a given open file object.

    This is a modified version of sncosmo.utils.download_file that
    downloads to an open file object instead of a cache directory.
    """

    from contextlib import closing
    from urllib.request import urlopen, Request, HTTPBasicAuthHandler
    from urllib.error import URLError, HTTPError
    from astropy.utils.console import ProgressBarOrSpinner
    from sncosmo import conf

    timeout = conf.remote_timeout
    download_block_size = 32768
    try:
        # Pretend to be a web browser (IE 6.0). Some servers that we download
        # from forbid access from programs.
        headers = {'User-Agent': 'Mozilla/5.0',
                   'Accept': ('text/html,application/xhtml+xml,'
                              'application/xml;q=0.9,*/*;q=0.8')}
        req = Request(remote_url, headers=headers)

        if credentials is not None:
            #            creds = base64.base64encode(f'{credentials.username}:{credentials.password}'.encode()).decode()
            #            req.add_header('Authorization', f'Basic {creds}')
            auth_handler = HTTPBasicAuthHandler()
            auth_handler.add_password(credentials.realm,
                                      uri=remote_url,
                                      user=credentials.username,
                                      passwd=credentials.password)
            opener = urllib.request.build_opener(auth_handler)
        else:
            opener = urllib.request.build_opener()

        #  with closing(urlopen(req, timeout=timeout)) as remote:
        with closing(opener.open(req, timeout=timeout)) as remote:

            # get size of remote if available (for use in progress bar)
            info = remote.info()
            size = None
            if 'Content-Length' in info:
                try:
                    size = int(info['Content-Length'])
                except ValueError:
                    pass

            dlmsg = "Downloading {0}".format(remote_url)
            with ProgressBarOrSpinner(size, dlmsg) as p:
                bytes_read = 0
                block = remote.read(download_block_size)
                while block:
                    target.write(block)
                    bytes_read += len(block)
                    p.update(bytes_read)
                    block = remote.read(download_block_size)

    # Append a more informative error message to HTTPErrors, URLErrors.
    except HTTPError as e:
        e.msg = "{}. requested URL: {!r}".format(e.msg, remote_url)
        raise
    except URLError as e:
        append_msg = (hasattr(e, 'reason') and hasattr(e.reason, 'errno') and
                      e.reason.errno == 8)
        if append_msg:
            msg = "{0}. requested URL: {1}".format(e.reason.strerror,
                                                   remote_url)
            e.reason.strerror = msg
            e.reason.args = (e.reason.errno, msg)
        raise e

    # This isn't supposed to happen, but occasionally a socket.timeout gets
    # through.  It's supposed to be caught in `urrlib2` and raised in this
    # way, but for some reason in mysterious circumstances it doesn't. So
    # we'll just re-raise it here instead.
    except socket.timeout as e:
        # add the requested URL to the message (normally just 'timed out')
        e.args = ('requested URL {!r} timed out'.format(remote_url),)
        raise URLError(e)


def download_file(remote_url, local_name, credentials=None):
    """
    Download a remote file to local path, unzipping if the URL ends in '.gz'.

    Parameters
    ----------
    remote_url : str
        The URL of the file to download

    local_name : str
        Absolute path filename of target file.

    Raises
    ------
    URLError
        Whenever there's a problem getting the remote file.
    """

    # ensure target directory exists
    dn = os.path.dirname(local_name)
    if not os.path.exists(dn):
        os.makedirs(dn)

    if remote_url.endswith(".gz"):
        import io
        import gzip

        buf = io.BytesIO()
        _download_file(remote_url, buf, credentials=credentials)
        buf.seek(0)
        f = gzip.GzipFile(fileobj=buf, mode='rb')

        with open(local_name, 'wb') as target:
            target.write(f.read())
        f.close()

    else:
        try:
            with open(local_name, 'wb') as target:
                _download_file(remote_url, target, credentials=credentials)
        except:  # noqa
            # in case of error downloading, remove file.
            if os.path.exists(local_name):
                os.remove(local_name)
            raise


def download_dir(remote_url, dirname, credentials=None):
    """
    Download a remote tar file to a local directory.

    Parameters
    ----------
    remote_url : str
        The URL of the file to download

    dirname : str
        Directory in which to place contents of tarfile. Created if it
        doesn't exist.

    Raises
    ------
    URLError (from urllib2 on PY2, urllib.request on PY3)
        Whenever there's a problem getting the remote file.
    """

    import io
    import tarfile

    if not os.path.exists(dirname):
        os.makedirs(dirname)

    mode = 'r:gz' if remote_url.endswith(".gz") else None

    # download file to buffer
    buf = io.BytesIO()
    _download_file(remote_url, buf, credentials=credentials)
    buf.seek(0)

    # create a tarfile with the buffer and extract
    tf = tarfile.open(fileobj=buf, mode=mode)
    tf.extractall(path=dirname)
    tf.close()
    buf.close()  # buf not closed when tf is closed.


# TODO: should inherit from DataMirror, I guess
class DataMirrorWithCredentials(object):
    """Lazy fetcher for remote data.

    When asked for local absolute path to a file or directory, DataMirror
    checks if the file or directory exists locally and, if so, returns it.

    If it doesn't exist, it first determines where to get it from.
    It first downloads the file ``{remote_root}/redirects.json`` and checks
    it for a redirect from ``{relative_path}`` to a full URL. If no redirect
    exists, it uses ``{remote_root}/{relative_path}`` as the URL.

    It downloads then downloads the URL to ``{rootdir}/{relative_path}``.

    For directories, ``.tar.gz`` is appended to the
    ``{relative_path}`` before the above is done and then the
    directory is unpacked locally.

    Parameters
    ----------
    rootdir : str or callable

        The local root directory, or a callable that returns the local root
        directory given no parameters. (The result of the call is cached.)
        Using a callable allows one to customize the discovery of the root
        directory (e.g., from a config file), and to defer that discovery
        until it is needed.

    remote_root : str
        Root URL of the remote server.
    """

    def __init__(self, rootdir, remote_root, credentials=None):
        if not remote_root.endswith('/'):
            remote_root = remote_root + '/'

        self._checked_rootdir = None
        self._rootdir = rootdir
        self._remote_root = remote_root

        self._redirects = None
        self._credentials = credentials

    def rootdir(self):
        """Return the path to the local data directory, ensuring that it
        exists"""

        if self._checked_rootdir is None:

            # If the supplied value is a string, use it. Otherwise
            # assume it is a callable that returns a string)
            rootdir = (self._rootdir
                       if isinstance(self._rootdir, str)
                       else self._rootdir())

            # Check existance
            if not os.path.isdir(rootdir):
                raise Exception("data directory {!r} not an existing "
                                "directory".format(rootdir))

            # Cache value for future calls
            self._checked_rootdir = rootdir

        return self._checked_rootdir

    def _fetch_redirects(self):
        import json
        # from urllib.request import urlopen
        if self._credentials is not None:
            from urllib.request import HTTPBasicAuthHandler
            auth_handler = HTTPBasicAuthHandler()
            auth_handler.add_password(realm=self._credentials.realm,
                                      uri=self._remote_root,
                                      user=self._credentials.username,
                                      passwd=self._credentials.password)
            opener = urllib.request.build_opener(auth_handler)
        else:
            opener = urllib.request.build_opener()

        # f = urlopen(self._remote_root + "redirects.json")
        f = opener.open(self._remote_root + "redirects.json")
        reader = codecs.getreader("utf-8")
        self._redirects = json.load(reader(f))
        f.close()

    def _get_url(self, remote_relpath):
        if self._redirects is None:
            self._fetch_redirects()

        if remote_relpath in self._redirects:
            return self._redirects[remote_relpath]
        else:
            return self._remote_root + remote_relpath

    def abspath(self, relpath, isdir=False):
        """Return absolute path to file or directory, ensuring that it exists.

        If ``isdir``, look for ``{relpath}.tar.gz`` on the remote server and
        unpackage it.

        Otherwise, just look for ``{relpath}``. If redirect points to a gz, it
        will be uncompressed."""

        abspath = os.path.join(self.rootdir(), relpath)

        if not os.path.exists(abspath):
            if isdir:
                url = self._get_url(relpath + ".tar.gz")

                # Download and unpack a directory.
                download_dir(url, os.path.dirname(abspath), credentials=self._credentials)

                # ensure that tarfile unpacked into the expected directory
                if not os.path.exists(abspath):
                    raise RuntimeError("Tarfile not unpacked into expected "
                                       "subdirectory. Please file an issue.")
            else:
                url = self._get_url(relpath)
                print(relpath, url)
                download_file(url, abspath, credentials=self._credentials)

        return abspath




#!/usr/bin/env python3

import logging

import numpy as np
import matplotlib.pyplot as plt

from lemaitre import bandpasses
from nacl.models import helpers
from nacl.models import salt2
from nacl.models.salt2 import lightcurves
from nacl.models.salt2 import spectra
from nacl import TrainingDataset


def load(filename='lemaitre.compressed.parquet',
         keep=['ZTF19aavhvkd', 'ZTF20abgfekk',
               '08D3dc', '07D2kc', '04D4ju',
               'ZTF18abnxhmc', 'ZTF19acsvqrx', 'ZTF20abkdpsy']):
    """
    """
    fl = bandpasses.get_filterlib()

    if 'hd5' in filename:
        tds = TrainingDataset.read_hdf(filename)
    else:
        tds = TrainingDataset.read_parquet(filename, filterlib=fl)

    if keep is not None:
        sne = tds.sn_data.sn_set.tolist()
        for k in keep:
            sne.remove(k)
        tds.kill_sne(sne)
        tds.compress()

    return tds


def test_deriv(model, pars, block_name='M0'):
    """
    """
    #       pars = model.init_pars()
    pars.fix()
    pars.release(block_name)

    J, Jn = helpers.check_grad(model, pars, dx=1.E-7)
    print(J.shape, Jn.shape)
    J = J.ravel()
    Jn = Jn.ravel()
    idx = (np.abs(J) > 0) | (np.abs(Jn) > 0)

    fig, axes = plt.subplots(nrows=2, ncols=2, figsize=(10,10))
    fig.suptitle(f'derivatives: {block_name}')
    axes[0,0].plot(J[idx], Jn[idx], 'k.')
    axes[0,0].set_xlabel('$\partial_{true}$')
    axes[0,0].set_ylabel('$\partial_{num}$')

    axes[0,1].plot((J[idx] - Jn[idx])/Jn[idx], 'k.')
    axes[0,1].set_xlabel('i')
    axes[0,1].set_ylabel('$(\partial_{true} - \partial_{num}) / \partial_{num}$')

    axes[1,0].plot(J[idx], (J[idx] - Jn[idx])/Jn[idx], 'k.')
    axes[1,0].set_xlabel('$\partial_{true}$')
    axes[1,0].set_ylabel('$(\partial_{num} - \partial_{true})/\partial_{num}$')

    axes[1,1].plot(J[idx] - Jn[idx], 'k.')
    axes[1,1].set_xlabel('$i$')
    axes[1,1].set_ylabel('$\partial_{true} - \partial_{num}$')

    return J, Jn


if __name__ == '__main__':
    tds = load()
    model = salt2.SALT2Like(tds)
    pars = model.init_pars()

    for block_name in ['X0', 'X1', 'col', 'tmax', 'M0', 'M1', 'CL', 'SpectrumRecalibration']:
        logging.info(f'testing: {block_name}')
        test_deriv(model, pars, block_name)


    pars.fix()
    pars.release('M0')
    J, Jn = helpers.check_grad(model, pars, dx=1.E-7)
    fig, axes = plt.subplots(nrows=1, ncols=3, figsize=(16,5),
                             sharex=True, sharey=True)
    fig.suptitle('M0')
    axes[0].imshow(J, aspect='auto', vmin=-0.1, vmax=0.1)
    axes[0].set_title('analytical')
    axes[1].imshow(Jn, aspect='auto', vmin=-0.1, vmax=0.1)
    axes[1].set_title('numerical')
    c = axes[2].imshow((J-Jn)/Jn, aspect='auto', vmin=-1., vmax=1.)
    plt.colorbar(c)
    axes[2].set_title('(J-Jn)/Jn')

    pars.fix()
    pars.release('M1')
    J, Jn = helpers.check_grad(model, pars)
    fig.suptitle('M1')
    fig, axes = plt.subplots(nrows=1, ncols=3, figsize=(16,5),
                             sharex=True, sharey=True)
    axes[0].imshow(J, aspect='auto', vmin=-0.1, vmax=0.1)
    axes[0].set_title('analytical')
    axes[1].imshow(Jn, aspect='auto', vmin=-0.1, vmax=0.1)
    axes[1].set_title('numerical')
    c = axes[2].imshow((J-Jn)/Jn, aspect='auto', vmin=-1., vmax=1.)
    plt.colorbar(c)
    axes[2].set_title('(J-Jn)/Jn')


# def test_kron(filename='compressed_lm2'):

#     tds = load(filename)
#     model = salt2.SALT2Like(tds)
#     pars = model.init_pars()
#     ll = lightcurves.LightcurveEvalUnit(model)
#     F = ll.meas_filter_projections
#     restframe_phase = (tds.lc_data.mjd - pars['tmax'].full[tds.lc_data.sn_index]) / (1. + tds.lc_data.z)
#     ph_basis = ll.model.basis.by
#     J = ph_basis.eval(restframe_phase)

#     return J, F



# def test_deriv(model, pars, block_name='M0'):
#     """
#     """
#     #       pars = model.init_pars()
#     pars.fix()
#     pars.release(block_name)

#     J, Jn = helpers.check_grad(model, pars)
#     print(J.shape, Jn.shape)
#     J = J.ravel()
#     Jn = Jn.ravel()
#     idx = (np.abs(J) > 0) | (np.abs(Jn) > 0)

#     fig, axes = plt.subplots(nrows=2, ncols=2, figsize=(10,10))
#     fig.suptitle(f'derivatives: {block_name}')
#     axes[0,0].plot(J[idx], Jn[idx], 'k.')
#     axes[0,0].set_xlabel('$\partial_{true}$')
#     axes[0,0].set_ylabel('$\partial_{num}$')

#     axes[1,1].plot(J[idx] - Jn[idx], 'k.')
#     axes[1,1].set_xlabel('$i$')
#     axes[1,1].set_ylabel('$\partial_{true} - \partial_{num}$')

#     axes[0,1].plot((J[idx] - Jn[idx])/J[idx], 'k.')
#     axes[0,1].set_xlabel('i')
#     axes[0,1].set_ylabel('$(\partial_{num} - \partial_{true}) / \partial_{true}$')

#     axes[1,0].plot(J[idx], J[idx] - Jn[idx], 'k.')
#     axes[1,0].set_xlabel('$\partial_{true}$')
#     axes[1,0].set_ylabel('$\partial_{num} - \partial_{true}$')

#     return J, Jn
